# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

import numpy as np

from ..._fiff._digitization import DigPoint, _ensure_fiducials_head
from ..._fiff.constants import FIFF
from ..._fiff.meas_info import create_info
from ..._fiff.pick import pick_info
from ...transforms import rotation3d_align_z_axis
from ...utils import _check_pandas_installed, warn

_supported_megs = ["neuromag306"]

_unit_dict = {
    "m": 1,
    "cm": 1e-2,
    "mm": 1e-3,
    "V": 1,
    "mV": 1e-3,
    "uV": 1e-6,
    "T": 1,
    "T/m": 1,
    "T/cm": 1e2,
}

NOINFO_WARNING = (
    "Importing FieldTrip data without an info dict from the "
    "original file. Channel locations, orientations and types "
    "will be incorrect. The imported data cannot be used for "
    "source analysis, channel interpolation etc."
)


def _validate_ft_struct(ft_struct):
    """Run validation checks on the ft_structure."""
    if isinstance(ft_struct, list):
        raise RuntimeError("Loading of data in cell arrays is not supported")


def _create_info(ft_struct, raw_info):
    """Create MNE info structure from a FieldTrip structure."""
    if raw_info is None:
        warn(NOINFO_WARNING)

    sfreq = _set_sfreq(ft_struct)
    ch_names = ft_struct["label"]
    if raw_info:
        info = raw_info.copy()
        missing_channels = set(ch_names) - set(info["ch_names"])
        if missing_channels:
            warn(
                "The following channels are present in the FieldTrip data "
                f"but cannot be found in the provided info: {missing_channels}.\n"
                "These channels will be removed from the resulting data!"
            )

            missing_chan_idx = [ch_names.index(ch) for ch in missing_channels]
            new_chs = [ch for ch in ch_names if ch not in missing_channels]
            ch_names = new_chs
            ft_struct["label"] = ch_names

            if "trial" in ft_struct:
                ft_struct["trial"] = _remove_missing_channels_from_trial(
                    ft_struct["trial"], missing_chan_idx
                )

            if "avg" in ft_struct:
                if ft_struct["avg"].ndim == 2:
                    ft_struct["avg"] = np.delete(
                        ft_struct["avg"], missing_chan_idx, axis=0
                    )

        with info._unlock():
            info["sfreq"] = sfreq
        ch_idx = [info["ch_names"].index(ch) for ch in ch_names]
        pick_info(info, ch_idx, copy=False)
    else:
        info = create_info(ch_names, sfreq)
        chs, dig = _create_info_chs_dig(ft_struct)
        with info._unlock(update_redundant=True):
            info.update(chs=chs, dig=dig)

    return info


def _remove_missing_channels_from_trial(trial, missing_chan_idx):
    if isinstance(trial, list):
        for idx_trial in range(len(trial)):
            trial[idx_trial] = _remove_missing_channels_from_trial(
                trial[idx_trial], missing_chan_idx
            )
    elif isinstance(trial, np.ndarray):
        if trial.ndim == 2:
            trial = np.delete(trial, missing_chan_idx, axis=0)
    else:
        raise ValueError(
            '"trial" field of the FieldTrip structure has an unknown format.'
        )

    return trial


def _create_info_chs_dig(ft_struct):
    """Create the chs info field from the FieldTrip structure."""
    all_channels = ft_struct["label"]
    ch_defaults = dict(
        coord_frame=FIFF.FIFFV_COORD_UNKNOWN,
        cal=1.0,
        range=1.0,
        unit_mul=FIFF.FIFF_UNITM_NONE,
        loc=np.array([0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1]),
        unit=FIFF.FIFF_UNIT_V,
    )
    try:
        elec = ft_struct["elec"]
    except KeyError:
        elec = None

    try:
        grad = ft_struct["grad"]
    except KeyError:
        grad = None

    if elec is None and grad is None:
        warn(
            "The supplied FieldTrip structure does not have an elec or grad "
            "field. No channel locations will extracted and the kind of "
            "channel might be inaccurate."
        )
    if "chanpos" not in (elec or grad or {"chanpos": None}):
        raise RuntimeError(
            "This file was created with an old version of FieldTrip. You can "
            "convert the data to the new version by loading it into FieldTrip "
            "and applying ft_selectdata with an empty cfg structure on it. "
            "Otherwise you can supply the Info field."
        )

    chs = list()
    dig = list()
    counter = 0
    for idx_chan, cur_channel_label in enumerate(all_channels):
        cur_ch = ch_defaults.copy()
        cur_ch["ch_name"] = cur_channel_label
        cur_ch["logno"] = idx_chan + 1
        cur_ch["scanno"] = idx_chan + 1
        if elec and cur_channel_label in elec["label"]:
            cur_ch = _process_channel_eeg(cur_ch, elec)
            assert cur_ch["coord_frame"] == FIFF.FIFFV_COORD_HEAD
            # Ref gets ident=0 and we don't have it, so start at 1
            counter += 1
            d = DigPoint(
                r=cur_ch["loc"][:3],
                coord_frame=FIFF.FIFFV_COORD_HEAD,
                kind=FIFF.FIFFV_POINT_EEG,
                ident=counter,
            )
            dig.append(d)
        elif grad and cur_channel_label in grad["label"]:
            cur_ch = _process_channel_meg(cur_ch, grad)
        else:
            if cur_channel_label.startswith("EOG"):
                cur_ch["kind"] = FIFF.FIFFV_EOG_CH
                cur_ch["coil_type"] = FIFF.FIFFV_COIL_EEG
            elif cur_channel_label.startswith("ECG"):
                cur_ch["kind"] = FIFF.FIFFV_ECG_CH
                cur_ch["coil_type"] = FIFF.FIFFV_COIL_EEG_BIPOLAR
            elif cur_channel_label.startswith("STI"):
                cur_ch["kind"] = FIFF.FIFFV_STIM_CH
                cur_ch["coil_type"] = FIFF.FIFFV_COIL_NONE
            else:
                warn(
                    f"Cannot guess the correct type of channel {cur_channel_label}. "
                    "Making it a MISC channel."
                )
                cur_ch["kind"] = FIFF.FIFFV_MISC_CH
                cur_ch["coil_type"] = FIFF.FIFFV_COIL_NONE

        chs.append(cur_ch)
    _ensure_fiducials_head(dig)

    return chs, dig


def _set_sfreq(ft_struct):
    """Set the sample frequency."""
    try:
        sfreq = ft_struct["fsample"]
    except KeyError:
        try:
            time = ft_struct["time"]
        except KeyError:
            raise ValueError("No Source for sfreq found")
        else:
            t1, t2 = float(time[0]), float(time[1])
            sfreq = 1 / (t2 - t1)
    try:
        sfreq = float(sfreq)
    except TypeError:
        warn(
            "FieldTrip structure contained multiple sample rates, trying the "
            f"first of:\n{sfreq} Hz"
        )
        sfreq = float(sfreq.ravel()[0])
    return sfreq


def _set_tmin(ft_struct):
    """Set the start time before the event in evoked data if possible."""
    times = ft_struct["time"]
    time_check = all(times[i][0] == times[i - 1][0] for i, x in enumerate(times))
    if time_check:
        tmin = times[0][0]
    else:
        raise RuntimeError(
            "Loading data with non-uniform times per epoch is not supported"
        )
    return tmin


def _create_events(ft_struct, trialinfo_column):
    """Create an event matrix from the FieldTrip structure."""
    if "trialinfo" not in ft_struct:
        return None

    event_type = ft_struct["trialinfo"]
    event_number = range(len(event_type))

    if trialinfo_column < 0:
        raise ValueError("trialinfo_column must be positive")

    available_ti_cols = 1
    if event_type.ndim == 2:
        available_ti_cols = event_type.shape[1]

    if trialinfo_column > (available_ti_cols - 1):
        raise ValueError(
            "trialinfo_column is higher than the amount of columns in trialinfo."
        )

    event_trans_val = np.zeros(len(event_type))

    if event_type.ndim == 2:
        event_type = event_type[:, trialinfo_column]

    events = (
        np.vstack([np.array(event_number), event_trans_val, event_type]).astype("int").T
    )

    return events


def _create_event_metadata(ft_struct):
    """Create event metadata from trialinfo."""
    pandas = _check_pandas_installed(strict=False)
    if not pandas:
        warn(
            "The Pandas library is not installed. Not returning the original "
            "trialinfo matrix as metadata."
        )
        return None

    metadata = pandas.DataFrame(ft_struct["trialinfo"])

    return metadata


def _process_channel_eeg(cur_ch, elec):
    """Convert EEG channel from FieldTrip to MNE.

    Parameters
    ----------
    cur_ch: dict
        Channel specific dictionary to populate.

    elec: dict
        elec dict as loaded from the FieldTrip structure

    Returns
    -------
    cur_ch: dict
        The original dict (cur_ch) with the added information
    """
    all_labels = np.asanyarray(elec["label"])
    chan_idx_in_elec = np.where(all_labels == cur_ch["ch_name"])[0][0]
    position = np.squeeze(elec["chanpos"][chan_idx_in_elec, :])
    # chanunit = elec['chanunit'][chan_idx_in_elec]  # not used/needed yet
    position_unit = elec["unit"]

    position = position * _unit_dict[position_unit]
    cur_ch["loc"] = np.hstack((position, np.zeros((9,))))
    cur_ch["unit"] = FIFF.FIFF_UNIT_V
    cur_ch["kind"] = FIFF.FIFFV_EEG_CH
    cur_ch["coil_type"] = FIFF.FIFFV_COIL_EEG
    cur_ch["coord_frame"] = FIFF.FIFFV_COORD_HEAD

    return cur_ch


def _process_channel_meg(cur_ch, grad):
    """Convert MEG channel from FieldTrip to MNE.

    Parameters
    ----------
    cur_ch: dict
        Channel specific dictionary to populate.

    grad: dict
        grad dict as loaded from the FieldTrip structure

    Returns
    -------
    dict: The original dict (cur_ch) with the added information
    """
    all_labels = np.asanyarray(grad["label"])
    chan_idx_in_grad = np.where(all_labels == cur_ch["ch_name"])[0][0]
    gradtype = grad["type"]
    chantype = grad["chantype"][chan_idx_in_grad]
    position_unit = grad["unit"]
    position = np.squeeze(grad["chanpos"][chan_idx_in_grad, :])
    position = position * _unit_dict[position_unit]

    if gradtype == "neuromag306" and "tra" in grad and "coilpos" in grad:
        # Try to regenerate original channel pos.
        idx_in_coilpos = np.where(grad["tra"][chan_idx_in_grad, :] != 0)[0]
        cur_coilpos = grad["coilpos"][idx_in_coilpos, :]
        cur_coilpos = cur_coilpos * _unit_dict[position_unit]
        cur_coilori = grad["coilori"][idx_in_coilpos, :]
        if chantype == "megmag":
            position = cur_coilpos[0] - 0.0003 * cur_coilori[0]
        if chantype == "megplanar":
            tmp_pos = cur_coilpos - 0.0003 * cur_coilori
            position = np.average(tmp_pos, axis=0)

    original_orientation = np.squeeze(grad["chanori"][chan_idx_in_grad, :])
    try:
        orientation = rotation3d_align_z_axis(original_orientation).T
    except AssertionError:
        orientation = np.eye(3)
    assert orientation.shape == (3, 3)
    orientation = orientation.flatten()
    # chanunit = grad['chanunit'][chan_idx_in_grad]  # not used/needed yet

    cur_ch["loc"] = np.hstack((position, orientation))
    cur_ch["kind"] = FIFF.FIFFV_MEG_CH
    if chantype == "megmag":
        cur_ch["coil_type"] = FIFF.FIFFV_COIL_POINT_MAGNETOMETER
        cur_ch["unit"] = FIFF.FIFF_UNIT_T
    elif chantype == "megplanar":
        cur_ch["coil_type"] = FIFF.FIFFV_COIL_VV_PLANAR_T1
        cur_ch["unit"] = FIFF.FIFF_UNIT_T_M
    elif chantype == "refmag":
        cur_ch["coil_type"] = FIFF.FIFFV_COIL_MAGNES_REF_MAG
        cur_ch["unit"] = FIFF.FIFF_UNIT_T
    elif chantype == "refgrad":
        cur_ch["coil_type"] = FIFF.FIFFV_COIL_MAGNES_REF_GRAD
        cur_ch["unit"] = FIFF.FIFF_UNIT_T
    elif chantype == "meggrad":
        cur_ch["coil_type"] = FIFF.FIFFV_COIL_AXIAL_GRAD_5CM
        cur_ch["unit"] = FIFF.FIFF_UNIT_T
    else:
        raise RuntimeError(f"Unexpected coil type: {chantype}.")

    cur_ch["coord_frame"] = FIFF.FIFFV_COORD_HEAD

    return cur_ch



# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

from ...utils import verbose
from ..utils import _data_path_doc, _download_mne_dataset, _get_version, _version_doc


@verbose
def data_path(
    path=None, force_update=False, update_path=True, download=True, *, verbose=None
):  # noqa: D103
    return _download_mne_dataset(
        name="fnirs_motor",
        processor="untar",
        path=path,
        force_update=force_update,
        update_path=update_path,
        download=download,
    )


data_path.__doc__ = _data_path_doc.format(
    name="fnirs_motor", conf="MNE_DATASETS_FNIRS_MOTOR_PATH"
)


def get_version():  # noqa: D103
    return _get_version("fnirs_motor")


get_version.__doc__ = _version_doc.format(name="fnirs_motor")



"""EGI module for conversion to FIF."""

# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

from .egi import read_raw_egi
from .egimff import read_evokeds_mff



# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

from ..utils._bunch import BunchConstNamed

FIFF = BunchConstNamed()

#
# FIFF version number in use
#
FIFF.FIFFC_MAJOR_VERSION = 1
FIFF.FIFFC_MINOR_VERSION = 4
FIFF.FIFFC_VERSION = FIFF.FIFFC_MAJOR_VERSION << 16 | FIFF.FIFFC_MINOR_VERSION

#
# Blocks
#
FIFF.FIFFB_ROOT = 999
FIFF.FIFFB_MEAS = 100
FIFF.FIFFB_MEAS_INFO = 101
FIFF.FIFFB_RAW_DATA = 102
FIFF.FIFFB_PROCESSED_DATA = 103
FIFF.FIFFB_EVOKED = 104
FIFF.FIFFB_ASPECT = 105
FIFF.FIFFB_SUBJECT = 106
FIFF.FIFFB_ISOTRAK = 107
FIFF.FIFFB_HPI_MEAS = 108  # HPI measurement
FIFF.FIFFB_HPI_RESULT = 109  # Result of a HPI fitting procedure
FIFF.FIFFB_HPI_COIL = 110  # Data acquired from one HPI coil
FIFF.FIFFB_PROJECT = 111
FIFF.FIFFB_CONTINUOUS_DATA = 112
FIFF.FIFFB_CH_INFO = 113  # Extra channel information
FIFF.FIFFB_VOID = 114
FIFF.FIFFB_EVENTS = 115
FIFF.FIFFB_INDEX = 116
FIFF.FIFFB_DACQ_PARS = 117
FIFF.FIFFB_REF = 118
FIFF.FIFFB_IAS_RAW_DATA = 119
FIFF.FIFFB_IAS_ASPECT = 120
FIFF.FIFFB_HPI_SUBSYSTEM = 121
# FIFF.FIFFB_PHANTOM_SUBSYSTEM  = 122
# FIFF.FIFFB_STATUS_SUBSYSTEM   = 123
FIFF.FIFFB_DEVICE = 124
FIFF.FIFFB_HELIUM = 125
FIFF.FIFFB_CHANNEL_INFO = 126

FIFF.FIFFB_SPHERE = 300  # Concentric sphere model related
FIFF.FIFFB_BEM = 310  # Boundary-element method
FIFF.FIFFB_BEM_SURF = 311  # Boundary-element method surfaces
FIFF.FIFFB_CONDUCTOR_MODEL = 312  # One conductor model definition
FIFF.FIFFB_PROJ = 313
FIFF.FIFFB_PROJ_ITEM = 314
FIFF.FIFFB_MRI = 200
FIFF.FIFFB_MRI_SET = 201
FIFF.FIFFB_MRI_SLICE = 202
FIFF.FIFFB_MRI_SCENERY = 203  # These are for writing unrelated 'slices'
FIFF.FIFFB_MRI_SCENE = 204  # Which are actually 3D scenes...
FIFF.FIFFB_MRI_SEG = 205  # MRI segmentation data
FIFF.FIFFB_MRI_SEG_REGION = 206  # One MRI segmentation region
FIFF.FIFFB_PROCESSING_HISTORY = 900
FIFF.FIFFB_PROCESSING_RECORD = 901

FIFF.FIFFB_DATA_CORRECTION = 500
FIFF.FIFFB_CHANNEL_DECOUPLER = 501
FIFF.FIFFB_SSS_INFO = 502
FIFF.FIFFB_SSS_CAL = 503
FIFF.FIFFB_SSS_ST_INFO = 504
FIFF.FIFFB_SSS_BASES = 505
FIFF.FIFFB_IAS = 510
#
# Of general interest
#
FIFF.FIFF_FILE_ID = 100
FIFF.FIFF_DIR_POINTER = 101
FIFF.FIFF_BLOCK_ID = 103
FIFF.FIFF_BLOCK_START = 104
FIFF.FIFF_BLOCK_END = 105
FIFF.FIFF_FREE_LIST = 106
FIFF.FIFF_FREE_BLOCK = 107
FIFF.FIFF_NOP = 108
FIFF.FIFF_PARENT_FILE_ID = 109
FIFF.FIFF_PARENT_BLOCK_ID = 110
FIFF.FIFF_BLOCK_NAME = 111
FIFF.FIFF_BLOCK_VERSION = 112
FIFF.FIFF_CREATOR = 113  # Program that created the file (string)
FIFF.FIFF_MODIFIER = 114  # Program that modified the file (string)
FIFF.FIFF_REF_ROLE = 115
FIFF.FIFF_REF_FILE_ID = 116
FIFF.FIFF_REF_FILE_NUM = 117
FIFF.FIFF_REF_FILE_NAME = 118
#
#  Megacq saves the parameters in these tags
#
FIFF.FIFF_DACQ_PARS = 150
FIFF.FIFF_DACQ_STIM = 151

FIFF.FIFF_DEVICE_TYPE = 152
FIFF.FIFF_DEVICE_MODEL = 153
FIFF.FIFF_DEVICE_SERIAL = 154
FIFF.FIFF_DEVICE_SITE = 155

FIFF.FIFF_HE_LEVEL_RAW = 156
FIFF.FIFF_HELIUM_LEVEL = 157
FIFF.FIFF_ORIG_FILE_GUID = 158
FIFF.FIFF_UTC_OFFSET = 159

FIFF.FIFF_NCHAN = 200
FIFF.FIFF_SFREQ = 201
FIFF.FIFF_DATA_PACK = 202
FIFF.FIFF_CH_INFO = 203
FIFF.FIFF_MEAS_DATE = 204
FIFF.FIFF_SUBJECT = 205
FIFF.FIFF_COMMENT = 206
FIFF.FIFF_NAVE = 207
FIFF.FIFF_FIRST_SAMPLE = 208  # The first sample of an epoch
FIFF.FIFF_LAST_SAMPLE = 209  # The last sample of an epoch
FIFF.FIFF_ASPECT_KIND = 210
FIFF.FIFF_REF_EVENT = 211
FIFF.FIFF_EXPERIMENTER = 212
FIFF.FIFF_DIG_POINT = 213
FIFF.FIFF_CH_POS = 214
FIFF.FIFF_HPI_SLOPES = 215  # HPI data
FIFF.FIFF_HPI_NCOIL = 216
FIFF.FIFF_REQ_EVENT = 217
FIFF.FIFF_REQ_LIMIT = 218
FIFF.FIFF_LOWPASS = 219
FIFF.FIFF_BAD_CHS = 220
FIFF.FIFF_ARTEF_REMOVAL = 221
FIFF.FIFF_COORD_TRANS = 222
FIFF.FIFF_HIGHPASS = 223
FIFF.FIFF_CH_CALS = 224  # This will not occur in new files
FIFF.FIFF_HPI_BAD_CHS = 225  # List of channels considered to be bad in hpi
FIFF.FIFF_HPI_CORR_COEFF = 226  # HPI curve fit correlations
FIFF.FIFF_EVENT_COMMENT = 227  # Comment about the events used in averaging
FIFF.FIFF_NO_SAMPLES = 228  # Number of samples in an epoch
FIFF.FIFF_FIRST_TIME = 229  # Time scale minimum

FIFF.FIFF_SUBAVE_SIZE = 230  # Size of a subaverage
FIFF.FIFF_SUBAVE_FIRST = 231  # The first epoch # contained in the subaverage
FIFF.FIFF_NAME = 233  # Intended to be a short name.
FIFF.FIFF_DESCRIPTION = FIFF.FIFF_COMMENT  # (Textual) Description of an object
FIFF.FIFF_DIG_STRING = 234  # String of digitized points
FIFF.FIFF_LINE_FREQ = 235  # Line frequency
FIFF.FIFF_GANTRY_ANGLE = 282  # Tilt angle of the gantry in degrees.

#
# HPI fitting program tags
#
FIFF.FIFF_HPI_COIL_FREQ = 236  # HPI coil excitation frequency
FIFF.FIFF_HPI_COIL_MOMENTS = (
    240  # Estimated moment vectors for the HPI coil magnetic dipoles
)
FIFF.FIFF_HPI_FIT_GOODNESS = 241  # Three floats indicating the goodness of fit
FIFF.FIFF_HPI_FIT_ACCEPT = 242  # Bitmask indicating acceptance (see below)
FIFF.FIFF_HPI_FIT_GOOD_LIMIT = 243  # Limit for the goodness-of-fit
FIFF.FIFF_HPI_FIT_DIST_LIMIT = 244  # Limit for the coil distance difference
FIFF.FIFF_HPI_COIL_NO = 245  # Coil number listed by HPI measurement
FIFF.FIFF_HPI_COILS_USED = (
    246  # List of coils finally used when the transformation was computed
)
FIFF.FIFF_HPI_DIGITIZATION_ORDER = (
    247  # Which Isotrak digitization point corresponds to each of the coils energized
)


#
# Tags used for storing channel info
#
FIFF.FIFF_CH_SCAN_NO = (
    250  # Channel scan number. Corresponds to fiffChInfoRec.scanNo field
)
FIFF.FIFF_CH_LOGICAL_NO = (
    251  # Channel logical number. Corresponds to fiffChInfoRec.logNo field
)
FIFF.FIFF_CH_KIND = 252  # Channel type. Corresponds to fiffChInfoRec.kind field"
FIFF.FIFF_CH_RANGE = (
    253  # Conversion from recorded number to (possibly virtual) voltage at the output"
)
FIFF.FIFF_CH_CAL = 254  # Calibration coefficient from output voltage to some real units
FIFF.FIFF_CH_LOC = 255  # Channel loc
FIFF.FIFF_CH_UNIT = 256  # Unit of the data
FIFF.FIFF_CH_UNIT_MUL = 257  # Unit multiplier exponent
FIFF.FIFF_CH_DACQ_NAME = 258  # Name of the channel in the data acquisition system. Corresponds to fiffChInfoRec.name.
FIFF.FIFF_CH_COIL_TYPE = 350  # Coil type in coil_def.dat
FIFF.FIFF_CH_COORD_FRAME = 351  # Coordinate frame (integer)

#
# Pointers
#
FIFF.FIFFV_NEXT_SEQ = 0
FIFF.FIFFV_NEXT_NONE = -1
#
# Channel types
#
FIFF.FIFFV_BIO_CH = 102
FIFF.FIFFV_MEG_CH = 1
FIFF.FIFFV_REF_MEG_CH = 301
FIFF.FIFFV_EEG_CH = 2
FIFF.FIFFV_MCG_CH = 201
FIFF.FIFFV_STIM_CH = 3
FIFF.FIFFV_EOG_CH = 202
FIFF.FIFFV_EMG_CH = 302
FIFF.FIFFV_ECG_CH = 402
FIFF.FIFFV_MISC_CH = 502
FIFF.FIFFV_RESP_CH = 602  # Respiration monitoring
FIFF.FIFFV_SEEG_CH = 802  # stereotactic EEG
FIFF.FIFFV_DBS_CH = 803  # deep brain stimulation
FIFF.FIFFV_SYST_CH = 900  # some system status information (on Triux systems only)
FIFF.FIFFV_ECOG_CH = 902
FIFF.FIFFV_IAS_CH = 910  # Internal Active Shielding data (maybe on Triux only)
FIFF.FIFFV_EXCI_CH = 920  # flux excitation channel used to be a stimulus channel
FIFF.FIFFV_DIPOLE_WAVE = 1000  # Dipole time curve (xplotter/xfit)
FIFF.FIFFV_GOODNESS_FIT = 1001  # Goodness of fit (xplotter/xfit)
FIFF.FIFFV_FNIRS_CH = 1100  # Functional near-infrared spectroscopy
FIFF.FIFFV_TEMPERATURE_CH = 1200  # Functional near-infrared spectroscopy
FIFF.FIFFV_GALVANIC_CH = 1300  # Galvanic skin response
FIFF.FIFFV_EYETRACK_CH = 1400  # Eye-tracking

_ch_kind_named = {
    key: key
    for key in (
        FIFF.FIFFV_BIO_CH,
        FIFF.FIFFV_MEG_CH,
        FIFF.FIFFV_REF_MEG_CH,
        FIFF.FIFFV_EEG_CH,
        FIFF.FIFFV_MCG_CH,
        FIFF.FIFFV_STIM_CH,
        FIFF.FIFFV_EOG_CH,
        FIFF.FIFFV_EMG_CH,
        FIFF.FIFFV_ECG_CH,
        FIFF.FIFFV_MISC_CH,
        FIFF.FIFFV_RESP_CH,
        FIFF.FIFFV_SEEG_CH,
        FIFF.FIFFV_DBS_CH,
        FIFF.FIFFV_SYST_CH,
        FIFF.FIFFV_ECOG_CH,
        FIFF.FIFFV_IAS_CH,
        FIFF.FIFFV_EXCI_CH,
        FIFF.FIFFV_DIPOLE_WAVE,
        FIFF.FIFFV_GOODNESS_FIT,
        FIFF.FIFFV_FNIRS_CH,
        FIFF.FIFFV_GALVANIC_CH,
        FIFF.FIFFV_TEMPERATURE_CH,
        FIFF.FIFFV_EYETRACK_CH,
    )
}

#
# Quaternion channels for head position monitoring
#
FIFF.FIFFV_QUAT_0 = 700  # Quaternion param q0 obsolete for unit quaternion
FIFF.FIFFV_QUAT_1 = 701  # Quaternion param q1 rotation
FIFF.FIFFV_QUAT_2 = 702  # Quaternion param q2 rotation
FIFF.FIFFV_QUAT_3 = 703  # Quaternion param q3 rotation
FIFF.FIFFV_QUAT_4 = 704  # Quaternion param q4 translation
FIFF.FIFFV_QUAT_5 = 705  # Quaternion param q5 translation
FIFF.FIFFV_QUAT_6 = 706  # Quaternion param q6 translation
FIFF.FIFFV_HPI_G = 707  # Goodness-of-fit in continuous hpi
FIFF.FIFFV_HPI_ERR = 708  # Estimation error in continuous hpi
FIFF.FIFFV_HPI_MOV = 709  # Estimated head movement speed in continuous hpi
#
# Coordinate frames
#
FIFF.FIFFV_COORD_UNKNOWN = 0
FIFF.FIFFV_COORD_DEVICE = 1
FIFF.FIFFV_COORD_ISOTRAK = 2
FIFF.FIFFV_COORD_HPI = 3
FIFF.FIFFV_COORD_HEAD = 4
FIFF.FIFFV_COORD_MRI = 5
FIFF.FIFFV_COORD_MRI_SLICE = 6
FIFF.FIFFV_COORD_MRI_DISPLAY = 7
FIFF.FIFFV_COORD_DICOM_DEVICE = 8
FIFF.FIFFV_COORD_IMAGING_DEVICE = 9
_coord_frame_named = {
    key: key
    for key in (
        FIFF.FIFFV_COORD_UNKNOWN,
        FIFF.FIFFV_COORD_DEVICE,
        FIFF.FIFFV_COORD_ISOTRAK,
        FIFF.FIFFV_COORD_HPI,
        FIFF.FIFFV_COORD_HEAD,
        FIFF.FIFFV_COORD_MRI,
        FIFF.FIFFV_COORD_MRI_SLICE,
        FIFF.FIFFV_COORD_MRI_DISPLAY,
        FIFF.FIFFV_COORD_DICOM_DEVICE,
        FIFF.FIFFV_COORD_IMAGING_DEVICE,
    )
}
#
# Needed for raw and evoked-response data
#
FIFF.FIFF_DATA_BUFFER = 300  # Buffer containing measurement data
FIFF.FIFF_DATA_SKIP = 301  # Data skip in buffers
FIFF.FIFF_EPOCH = 302  # Buffer containing one epoch and channel
FIFF.FIFF_DATA_SKIP_SAMP = 303  # Data skip in samples

#
# Info on subject
#
FIFF.FIFF_SUBJ_ID = 400  # Subject ID
FIFF.FIFF_SUBJ_FIRST_NAME = 401  # First name of the subject
FIFF.FIFF_SUBJ_MIDDLE_NAME = 402  # Middle name of the subject
FIFF.FIFF_SUBJ_LAST_NAME = 403  # Last name of the subject
FIFF.FIFF_SUBJ_BIRTH_DAY = 404  # Birthday of the subject
FIFF.FIFF_SUBJ_SEX = 405  # Sex of the subject
FIFF.FIFF_SUBJ_HAND = 406  # Handedness of the subject
FIFF.FIFF_SUBJ_WEIGHT = 407  # Weight of the subject in kg
FIFF.FIFF_SUBJ_HEIGHT = 408  # Height of the subject in m
FIFF.FIFF_SUBJ_COMMENT = 409  # Comment about the subject
FIFF.FIFF_SUBJ_HIS_ID = 410  # ID used in the Hospital Information System

FIFF.FIFFV_SUBJ_HAND_RIGHT = 1  # Righthanded
FIFF.FIFFV_SUBJ_HAND_LEFT = 2  # Lefthanded
FIFF.FIFFV_SUBJ_HAND_AMBI = 3  # Ambidextrous

FIFF.FIFFV_SUBJ_SEX_UNKNOWN = 0  # Unknown gender
FIFF.FIFFV_SUBJ_SEX_MALE = 1  # Male
FIFF.FIFFV_SUBJ_SEX_FEMALE = 2  # Female

FIFF.FIFF_PROJ_ID = 500
FIFF.FIFF_PROJ_NAME = 501
FIFF.FIFF_PROJ_AIM = 502
FIFF.FIFF_PROJ_PERSONS = 503
FIFF.FIFF_PROJ_COMMENT = 504

FIFF.FIFF_EVENT_CHANNELS = 600  # Event channel numbers
FIFF.FIFF_EVENT_LIST = 601  # List of events (integers: <sample before after>
FIFF.FIFF_EVENT_CHANNEL = 602  # Event channel
FIFF.FIFF_EVENT_BITS = 603  # Event bits array

#
# Tags used in saving SQUID characteristics etc.
#
FIFF.FIFF_SQUID_BIAS = 701
FIFF.FIFF_SQUID_OFFSET = 702
FIFF.FIFF_SQUID_GATE = 703
#
# Aspect values used to save characteristic curves of SQUIDs. (mjk)
#
FIFF.FIFFV_ASPECT_IFII_LOW = 1100
FIFF.FIFFV_ASPECT_IFII_HIGH = 1101
FIFF.FIFFV_ASPECT_GATE = 1102

#
# Values for file references
#
FIFF.FIFFV_ROLE_PREV_FILE = 1
FIFF.FIFFV_ROLE_NEXT_FILE = 2

#
# References
#
FIFF.FIFF_REF_PATH = 1101

#
# Different aspects of data
#
FIFF.FIFFV_ASPECT_AVERAGE = 100  # Normal average of epochs
FIFF.FIFFV_ASPECT_STD_ERR = 101  # Std. error of mean
FIFF.FIFFV_ASPECT_SINGLE = 102  # Single epoch cut out from the continuous data
FIFF.FIFFV_ASPECT_SUBAVERAGE = 103  # Partial average (subaverage)
FIFF.FIFFV_ASPECT_ALTAVERAGE = 104  # Alternating subaverage
FIFF.FIFFV_ASPECT_SAMPLE = 105  # A sample cut out by graph
FIFF.FIFFV_ASPECT_POWER_DENSITY = 106  # Power density spectrum
FIFF.FIFFV_ASPECT_DIPOLE_WAVE = 200  # Dipole amplitude curve

#
# BEM surface IDs
#
FIFF.FIFFV_BEM_SURF_ID_UNKNOWN = -1
FIFF.FIFFV_BEM_SURF_ID_NOT_KNOWN = 0
FIFF.FIFFV_BEM_SURF_ID_BRAIN = 1
FIFF.FIFFV_BEM_SURF_ID_CSF = 2
FIFF.FIFFV_BEM_SURF_ID_SKULL = 3
FIFF.FIFFV_BEM_SURF_ID_HEAD = 4

FIFF.FIFF_SPHERE_ORIGIN = 3001
FIFF.FIFF_SPHERE_RADIUS = 3002

FIFF.FIFF_BEM_SURF_ID = 3101  # int    surface number
FIFF.FIFF_BEM_SURF_NAME = 3102  # string surface name
FIFF.FIFF_BEM_SURF_NNODE = 3103  # int    number of nodes on a surface
FIFF.FIFF_BEM_SURF_NTRI = 3104  # int     number of triangles on a surface
FIFF.FIFF_BEM_SURF_NODES = 3105  # float  surface nodes (nnode,3)
FIFF.FIFF_BEM_SURF_TRIANGLES = 3106  # int    surface triangles (ntri,3)
FIFF.FIFF_BEM_SURF_NORMALS = 3107  # float  surface node normal unit vectors

FIFF.FIFF_BEM_POT_SOLUTION = 3110  # float ** The solution matrix
FIFF.FIFF_BEM_APPROX = 3111  # int    approximation method, see below
FIFF.FIFF_BEM_COORD_FRAME = 3112  # The coordinate frame of the model
FIFF.FIFF_BEM_SIGMA = 3113  # Conductivity of a compartment
FIFF.FIFFV_BEM_APPROX_CONST = 1  # The constant potential approach
FIFF.FIFFV_BEM_APPROX_LINEAR = 2  # The linear potential approach

#
# More of those defined in MNE
#
FIFF.FIFFV_MNE_SURF_UNKNOWN = -1
FIFF.FIFFV_MNE_SURF_LEFT_HEMI = 101
FIFF.FIFFV_MNE_SURF_RIGHT_HEMI = 102
FIFF.FIFFV_MNE_SURF_MEG_HELMET = 201  # Use this irrespective of the system
#
#   These relate to the Isotrak data (enum(point))
#
FIFF.FIFFV_POINT_CARDINAL = 1
FIFF.FIFFV_POINT_HPI = 2
FIFF.FIFFV_POINT_EEG = 3
FIFF.FIFFV_POINT_ECG = FIFF.FIFFV_POINT_EEG
FIFF.FIFFV_POINT_EXTRA = 4
FIFF.FIFFV_POINT_HEAD = 5  # Point on the surface of the head
_dig_kind_named = {
    key: key
    for key in (
        FIFF.FIFFV_POINT_CARDINAL,
        FIFF.FIFFV_POINT_HPI,
        FIFF.FIFFV_POINT_EEG,
        FIFF.FIFFV_POINT_EXTRA,
        FIFF.FIFFV_POINT_HEAD,
    )
}
#
# Cardinal point types (enum(cardinal_point))
#
FIFF.FIFFV_POINT_LPA = 1
FIFF.FIFFV_POINT_NASION = 2
FIFF.FIFFV_POINT_RPA = 3
FIFF.FIFFV_POINT_INION = 4
_dig_cardinal_named = {
    key: key
    for key in (
        FIFF.FIFFV_POINT_LPA,
        FIFF.FIFFV_POINT_NASION,
        FIFF.FIFFV_POINT_RPA,
        FIFF.FIFFV_POINT_INION,
    )
}
#
#   SSP
#
FIFF.FIFF_PROJ_ITEM_KIND = 3411
FIFF.FIFF_PROJ_ITEM_TIME = 3412
FIFF.FIFF_PROJ_ITEM_NVEC = 3414
FIFF.FIFF_PROJ_ITEM_VECTORS = 3415
FIFF.FIFF_PROJ_ITEM_DEFINITION = 3416
FIFF.FIFF_PROJ_ITEM_CH_NAME_LIST = 3417
#   XPlotter
FIFF.FIFF_XPLOTTER_LAYOUT = 3501  # string - "Xplotter layout tag"
#
#   MRIs
#
FIFF.FIFF_MRI_SOURCE_PATH = FIFF.FIFF_REF_PATH
FIFF.FIFF_MRI_SOURCE_FORMAT = 2002
FIFF.FIFF_MRI_PIXEL_ENCODING = 2003
FIFF.FIFF_MRI_PIXEL_DATA_OFFSET = 2004
FIFF.FIFF_MRI_PIXEL_SCALE = 2005
FIFF.FIFF_MRI_PIXEL_DATA = 2006
FIFF.FIFF_MRI_PIXEL_OVERLAY_ENCODING = 2007
FIFF.FIFF_MRI_PIXEL_OVERLAY_DATA = 2008
FIFF.FIFF_MRI_BOUNDING_BOX = 2009
FIFF.FIFF_MRI_WIDTH = 2010
FIFF.FIFF_MRI_WIDTH_M = 2011
FIFF.FIFF_MRI_HEIGHT = 2012
FIFF.FIFF_MRI_HEIGHT_M = 2013
FIFF.FIFF_MRI_DEPTH = 2014
FIFF.FIFF_MRI_DEPTH_M = 2015
FIFF.FIFF_MRI_THICKNESS = 2016
FIFF.FIFF_MRI_SCENE_AIM = 2017
FIFF.FIFF_MRI_ORIG_SOURCE_PATH = 2020
FIFF.FIFF_MRI_ORIG_SOURCE_FORMAT = 2021
FIFF.FIFF_MRI_ORIG_PIXEL_ENCODING = 2022
FIFF.FIFF_MRI_ORIG_PIXEL_DATA_OFFSET = 2023
FIFF.FIFF_MRI_VOXEL_DATA = 2030
FIFF.FIFF_MRI_VOXEL_ENCODING = 2031
FIFF.FIFF_MRI_MRILAB_SETUP = 2100
FIFF.FIFF_MRI_SEG_REGION_ID = 2200
#
FIFF.FIFFV_MRI_PIXEL_UNKNOWN = 0
FIFF.FIFFV_MRI_PIXEL_BYTE = 1
FIFF.FIFFV_MRI_PIXEL_WORD = 2
FIFF.FIFFV_MRI_PIXEL_SWAP_WORD = 3
FIFF.FIFFV_MRI_PIXEL_FLOAT = 4
FIFF.FIFFV_MRI_PIXEL_BYTE_INDEXED_COLOR = 5
FIFF.FIFFV_MRI_PIXEL_BYTE_RGB_COLOR = 6
FIFF.FIFFV_MRI_PIXEL_BYTE_RLE_RGB_COLOR = 7
FIFF.FIFFV_MRI_PIXEL_BIT_RLE = 8
#
#   These are the MNE fiff definitions (range 350-390 reserved for MNE)
#
FIFF.FIFFB_MNE = 350
FIFF.FIFFB_MNE_SOURCE_SPACE = 351
FIFF.FIFFB_MNE_FORWARD_SOLUTION = 352
FIFF.FIFFB_MNE_PARENT_MRI_FILE = 353
FIFF.FIFFB_MNE_PARENT_MEAS_FILE = 354
FIFF.FIFFB_MNE_COV = 355
FIFF.FIFFB_MNE_INVERSE_SOLUTION = 356
FIFF.FIFFB_MNE_NAMED_MATRIX = 357
FIFF.FIFFB_MNE_ENV = 358
FIFF.FIFFB_MNE_BAD_CHANNELS = 359
FIFF.FIFFB_MNE_VERTEX_MAP = 360
FIFF.FIFFB_MNE_EVENTS = 361
FIFF.FIFFB_MNE_MORPH_MAP = 362
FIFF.FIFFB_MNE_SURFACE_MAP = 363
FIFF.FIFFB_MNE_SURFACE_MAP_GROUP = 364

#
# CTF compensation data
#
FIFF.FIFFB_MNE_CTF_COMP = 370
FIFF.FIFFB_MNE_CTF_COMP_DATA = 371
FIFF.FIFFB_MNE_DERIVATIONS = 372

FIFF.FIFFB_MNE_EPOCHS = 373
FIFF.FIFFB_MNE_ICA = 374
#
# Fiff tags associated with MNE computations (3500...)
#
#
# 3500... Bookkeeping
#
FIFF.FIFF_MNE_ROW_NAMES = 3502
FIFF.FIFF_MNE_COL_NAMES = 3503
FIFF.FIFF_MNE_NROW = 3504
FIFF.FIFF_MNE_NCOL = 3505
FIFF.FIFF_MNE_COORD_FRAME = 3506  # Coordinate frame employed. Defaults:
#  FIFFB_MNE_SOURCE_SPACE       FIFFV_COORD_MRI
#  FIFFB_MNE_FORWARD_SOLUTION   FIFFV_COORD_HEAD
#  FIFFB_MNE_INVERSE_SOLUTION   FIFFV_COORD_HEAD
FIFF.FIFF_MNE_CH_NAME_LIST = 3507
FIFF.FIFF_MNE_FILE_NAME = (
    3508  # This removes the collision with fiff_file.h (used to be 3501)
)
#
# 3510... 3590... Source space or surface
#
FIFF.FIFF_MNE_SOURCE_SPACE_POINTS = 3510  # The vertices
FIFF.FIFF_MNE_SOURCE_SPACE_NORMALS = 3511  # The vertex normals
FIFF.FIFF_MNE_SOURCE_SPACE_NPOINTS = 3512  # How many vertices
FIFF.FIFF_MNE_SOURCE_SPACE_SELECTION = 3513  # Which are selected to the source space
FIFF.FIFF_MNE_SOURCE_SPACE_NUSE = 3514  # How many are in use
FIFF.FIFF_MNE_SOURCE_SPACE_NEAREST = (
    3515  # Nearest source space vertex for all vertices
)
FIFF.FIFF_MNE_SOURCE_SPACE_NEAREST_DIST = (
    3516  # Distance to the Nearest source space vertex for all vertices
)
FIFF.FIFF_MNE_SOURCE_SPACE_ID = 3517  # Identifier
FIFF.FIFF_MNE_SOURCE_SPACE_TYPE = 3518  # Surface or volume
FIFF.FIFF_MNE_SOURCE_SPACE_VERTICES = 3519  # List of vertices (zero based)

FIFF.FIFF_MNE_SOURCE_SPACE_VOXEL_DIMS = (
    3596  # Voxel space dimensions in a volume source space
)
FIFF.FIFF_MNE_SOURCE_SPACE_INTERPOLATOR = (
    3597  # Matrix to interpolate a volume source space into a mri volume
)
FIFF.FIFF_MNE_SOURCE_SPACE_MRI_FILE = 3598  # MRI file used in the interpolation

FIFF.FIFF_MNE_SOURCE_SPACE_NTRI = 3590  # Number of triangles
FIFF.FIFF_MNE_SOURCE_SPACE_TRIANGLES = 3591  # The triangulation
FIFF.FIFF_MNE_SOURCE_SPACE_NUSE_TRI = (
    3592  # Number of triangles corresponding to the number of vertices in use
)
FIFF.FIFF_MNE_SOURCE_SPACE_USE_TRIANGLES = (
    3593  # The triangulation of the used vertices in the source space
)
FIFF.FIFF_MNE_SOURCE_SPACE_NNEIGHBORS = 3594  # Number of neighbors for each source space point (used for volume source spaces)
FIFF.FIFF_MNE_SOURCE_SPACE_NEIGHBORS = (
    3595  # Neighbors for each source space point (used for volume source spaces)
)

FIFF.FIFF_MNE_SOURCE_SPACE_DIST = (
    3599  # Distances between vertices in use (along the surface)
)
FIFF.FIFF_MNE_SOURCE_SPACE_DIST_LIMIT = (
    3600  # If distance is above this limit (in the volume) it has not been calculated
)

FIFF.FIFF_MNE_SURFACE_MAP_DATA = 3610  # Surface map data
FIFF.FIFF_MNE_SURFACE_MAP_KIND = 3611  # Type of map

#
# 3520... Forward solution
#
FIFF.FIFF_MNE_FORWARD_SOLUTION = 3520
FIFF.FIFF_MNE_SOURCE_ORIENTATION = 3521  # Fixed or free
FIFF.FIFF_MNE_INCLUDED_METHODS = 3522
FIFF.FIFF_MNE_FORWARD_SOLUTION_GRAD = 3523
#
# 3530... Covariance matrix
#
FIFF.FIFF_MNE_COV_KIND = 3530  # What kind of a covariance matrix
FIFF.FIFF_MNE_COV_DIM = 3531  # Matrix dimension
FIFF.FIFF_MNE_COV = 3532  # Full matrix in packed representation (lower triangle)
FIFF.FIFF_MNE_COV_DIAG = 3533  # Diagonal matrix
FIFF.FIFF_MNE_COV_EIGENVALUES = 3534  # Eigenvalues and eigenvectors of the above
FIFF.FIFF_MNE_COV_EIGENVECTORS = 3535
FIFF.FIFF_MNE_COV_NFREE = 3536  # Number of degrees of freedom
FIFF.FIFF_MNE_COV_METHOD = 3537  # The estimator used
FIFF.FIFF_MNE_COV_SCORE = 3538  # Negative log-likelihood

#
# 3540... Inverse operator
#
# We store the inverse operator as the eigenleads, eigenfields,
# and weights
#
FIFF.FIFF_MNE_INVERSE_LEADS = 3540  # The eigenleads
FIFF.FIFF_MNE_INVERSE_LEADS_WEIGHTED = (
    3546  # The eigenleads (already weighted with R^0.5)
)
FIFF.FIFF_MNE_INVERSE_FIELDS = 3541  # The eigenfields
FIFF.FIFF_MNE_INVERSE_SING = 3542  # The singular values
FIFF.FIFF_MNE_PRIORS_USED = (
    3543  # Which kind of priors have been used for the source covariance matrix
)
FIFF.FIFF_MNE_INVERSE_FULL = 3544  # Inverse operator as one matrix
# This matrix includes the whitening operator as well
# The regularization is applied
FIFF.FIFF_MNE_INVERSE_SOURCE_ORIENTATIONS = (
    3545  # Contains the orientation of one source per row
)
# The source orientations must be expressed in the coordinate system
# given by FIFF_MNE_COORD_FRAME
FIFF.FIFF_MNE_INVERSE_SOURCE_UNIT = 3547  # Are the sources given in Am or Am/m^2 ?
#
# 3550... Saved environment info
#
FIFF.FIFF_MNE_ENV_WORKING_DIR = 3550  # Working directory where the file was created
FIFF.FIFF_MNE_ENV_COMMAND_LINE = 3551  # The command used to create the file
FIFF.FIFF_MNE_EXTERNAL_BIG_ENDIAN = (
    3552  # Reference to an external binary file (big-endian) */
)
FIFF.FIFF_MNE_EXTERNAL_LITTLE_ENDIAN = (
    3553  # Reference to an external binary file (little-endian) */
)
#
# 3560... Miscellaneous
#
FIFF.FIFF_MNE_PROJ_ITEM_ACTIVE = 3560  # Is this projection item active?
FIFF.FIFF_MNE_EVENT_LIST = 3561  # An event list (for STI101 / STI 014)
FIFF.FIFF_MNE_HEMI = 3562  # Hemisphere association for general purposes
FIFF.FIFF_MNE_DATA_SKIP_NOP = 3563  # A data skip turned off in the raw data
FIFF.FIFF_MNE_ORIG_CH_INFO = 3564  # Channel information before any changes
FIFF.FIFF_MNE_EVENT_TRIGGER_MASK = 3565  # Mask applied to the trigger channel values
FIFF.FIFF_MNE_EVENT_COMMENTS = 3566  # Event comments merged into one long string
FIFF.FIFF_MNE_CUSTOM_REF = 3567  # Whether a custom reference was applied to the data
FIFF.FIFF_MNE_BASELINE_MIN = 3568  # Time of baseline beginning
FIFF.FIFF_MNE_BASELINE_MAX = 3569  # Time of baseline end
#
# 3570... Morphing maps
#
FIFF.FIFF_MNE_MORPH_MAP = 3570  # Mapping of closest vertices on the sphere
FIFF.FIFF_MNE_MORPH_MAP_FROM = 3571  # Which subject is this map from
FIFF.FIFF_MNE_MORPH_MAP_TO = 3572  # Which subject is this map to
#
# 3580... CTF compensation data
#
FIFF.FIFF_MNE_CTF_COMP_KIND = 3580  # What kind of compensation
FIFF.FIFF_MNE_CTF_COMP_DATA = 3581  # The compensation data itself
FIFF.FIFF_MNE_CTF_COMP_CALIBRATED = 3582  # Are the coefficients calibrated?

FIFF.FIFF_MNE_DERIVATION_DATA = (
    3585  # Used to store information about EEG and other derivations
)
#
# 3601... values associated with ICA decomposition
#
FIFF.FIFF_MNE_ICA_INTERFACE_PARAMS = 3601  # ICA interface parameters
FIFF.FIFF_MNE_ICA_CHANNEL_NAMES = 3602  # ICA channel names
FIFF.FIFF_MNE_ICA_WHITENER = 3603  # ICA whitener
FIFF.FIFF_MNE_ICA_PCA_COMPONENTS = 3604  # PCA components
FIFF.FIFF_MNE_ICA_PCA_EXPLAINED_VAR = 3605  # PCA explained variance
FIFF.FIFF_MNE_ICA_PCA_MEAN = 3606  # PCA mean
FIFF.FIFF_MNE_ICA_MATRIX = 3607  # ICA unmixing matrix
FIFF.FIFF_MNE_ICA_BADS = 3608  # ICA bad sources
FIFF.FIFF_MNE_ICA_MISC_PARAMS = 3609  # ICA misc params
#
# Miscellaneous
#
FIFF.FIFF_MNE_KIT_SYSTEM_ID = 3612  # Unique ID assigned to KIT systems
#
# Maxfilter tags
#
FIFF.FIFF_SSS_FRAME = 263
FIFF.FIFF_SSS_JOB = 264
FIFF.FIFF_SSS_ORIGIN = 265
FIFF.FIFF_SSS_ORD_IN = 266
FIFF.FIFF_SSS_ORD_OUT = 267
FIFF.FIFF_SSS_NMAG = 268
FIFF.FIFF_SSS_COMPONENTS = 269
FIFF.FIFF_SSS_CAL_CHANS = 270
FIFF.FIFF_SSS_CAL_CORRS = 271
FIFF.FIFF_SSS_ST_CORR = 272
FIFF.FIFF_SSS_NFREE = 278
FIFF.FIFF_SSS_ST_LENGTH = 279
FIFF.FIFF_DECOUPLER_MATRIX = 800
#
# Fiff values associated with MNE computations
#
FIFF.FIFFV_MNE_UNKNOWN_ORI = 0
FIFF.FIFFV_MNE_FIXED_ORI = 1
FIFF.FIFFV_MNE_FREE_ORI = 2

FIFF.FIFFV_MNE_MEG = 1
FIFF.FIFFV_MNE_EEG = 2
FIFF.FIFFV_MNE_MEG_EEG = 3

FIFF.FIFFV_MNE_PRIORS_NONE = 0
FIFF.FIFFV_MNE_PRIORS_DEPTH = 1
FIFF.FIFFV_MNE_PRIORS_LORETA = 2
FIFF.FIFFV_MNE_PRIORS_SULCI = 3

FIFF.FIFFV_MNE_UNKNOWN_COV = 0
FIFF.FIFFV_MNE_SENSOR_COV = 1
FIFF.FIFFV_MNE_NOISE_COV = 1  # This is what it should have been called
FIFF.FIFFV_MNE_SOURCE_COV = 2
FIFF.FIFFV_MNE_FMRI_PRIOR_COV = 3
FIFF.FIFFV_MNE_SIGNAL_COV = 4  # This will be potentially employed in beamformers
FIFF.FIFFV_MNE_DEPTH_PRIOR_COV = 5  # The depth weighting prior
FIFF.FIFFV_MNE_ORIENT_PRIOR_COV = 6  # The orientation prior

#
# Output map types
#
FIFF.FIFFV_MNE_MAP_UNKNOWN = -1  # Unspecified
FIFF.FIFFV_MNE_MAP_SCALAR_CURRENT = 1  # Scalar current value
FIFF.FIFFV_MNE_MAP_SCALAR_CURRENT_SIZE = 2  # Absolute value of the above
FIFF.FIFFV_MNE_MAP_VECTOR_CURRENT = 3  # Current vector components
FIFF.FIFFV_MNE_MAP_VECTOR_CURRENT_SIZE = 4  # Vector current size
FIFF.FIFFV_MNE_MAP_T_STAT = 5  # Student's t statistic
FIFF.FIFFV_MNE_MAP_F_STAT = 6  # F statistic
FIFF.FIFFV_MNE_MAP_F_STAT_SQRT = 7  # Square root of the F statistic
FIFF.FIFFV_MNE_MAP_CHI2_STAT = 8  # (Approximate) chi^2 statistic
FIFF.FIFFV_MNE_MAP_CHI2_STAT_SQRT = (
    9  # Square root of the (approximate) chi^2 statistic
)
FIFF.FIFFV_MNE_MAP_SCALAR_CURRENT_NOISE = 10  # Current noise approximation (scalar)
FIFF.FIFFV_MNE_MAP_VECTOR_CURRENT_NOISE = 11  # Current noise approximation (vector)
#
# Source space types (values of FIFF_MNE_SOURCE_SPACE_TYPE)
#
FIFF.FIFFV_MNE_SPACE_UNKNOWN = -1
FIFF.FIFFV_MNE_SPACE_SURFACE = 1
FIFF.FIFFV_MNE_SPACE_VOLUME = 2
FIFF.FIFFV_MNE_SPACE_DISCRETE = 3
#
# Covariance matrix channel classification
#
FIFF.FIFFV_MNE_COV_CH_UNKNOWN = -1  # No idea
FIFF.FIFFV_MNE_COV_CH_MEG_MAG = 0  # Axial gradiometer or magnetometer [T]
FIFF.FIFFV_MNE_COV_CH_MEG_GRAD = 1  # Planar gradiometer [T/m]
FIFF.FIFFV_MNE_COV_CH_EEG = 2  # EEG [V]
#
# Projection item kinds
#
FIFF.FIFFV_PROJ_ITEM_NONE = 0
FIFF.FIFFV_PROJ_ITEM_FIELD = 1
FIFF.FIFFV_PROJ_ITEM_DIP_FIX = 2
FIFF.FIFFV_PROJ_ITEM_DIP_ROT = 3
FIFF.FIFFV_PROJ_ITEM_HOMOG_GRAD = 4
FIFF.FIFFV_PROJ_ITEM_HOMOG_FIELD = 5
FIFF.FIFFV_PROJ_ITEM_EEG_AVREF = (
    10  # Linear projection related to EEG average reference
)
FIFF.FIFFV_MNE_PROJ_ITEM_EEG_AVREF = (
    FIFF.FIFFV_PROJ_ITEM_EEG_AVREF
)  # backward compat alias
#
# Custom EEG references
#
FIFF.FIFFV_MNE_CUSTOM_REF_OFF = 0
FIFF.FIFFV_MNE_CUSTOM_REF_ON = 1
FIFF.FIFFV_MNE_CUSTOM_REF_CSD = 2
#
# SSS job options
#
FIFF.FIFFV_SSS_JOB_NOTHING = 0  # No SSS, just copy input to output
FIFF.FIFFV_SSS_JOB_CTC = 1  # No SSS, only cross-talk correction
FIFF.FIFFV_SSS_JOB_FILTER = 2  # Spatial maxwell filtering
FIFF.FIFFV_SSS_JOB_VIRT = 3  # Transform data to another sensor array
FIFF.FIFFV_SSS_JOB_HEAD_POS = 4  # Estimate head positions, no SSS
FIFF.FIFFV_SSS_JOB_MOVEC_FIT = 5  # Estimate and compensate head movement
FIFF.FIFFV_SSS_JOB_MOVEC_QUA = (
    6  # Compensate head movement from previously estimated head positions
)
FIFF.FIFFV_SSS_JOB_REC_ALL = 7  # Reconstruct inside and outside signals
FIFF.FIFFV_SSS_JOB_REC_IN = 8  # Reconstruct inside signals
FIFF.FIFFV_SSS_JOB_REC_OUT = 9  # Reconstruct outside signals
FIFF.FIFFV_SSS_JOB_ST = 10  # Spatio-temporal maxwell filtering
FIFF.FIFFV_SSS_JOB_TPROJ = 11  # Temporal projection, no SSS
FIFF.FIFFV_SSS_JOB_XSSS = 12  # Cross-validation SSS
FIFF.FIFFV_SSS_JOB_XSUB = 13  # Cross-validation subtraction, no SSS
FIFF.FIFFV_SSS_JOB_XWAV = 14  # Cross-validation noise waveforms
FIFF.FIFFV_SSS_JOB_NCOV = 15  # Noise covariance estimation
FIFF.FIFFV_SSS_JOB_SCOV = 16  # SSS sample covariance estimation
# }

#
# Additional coordinate frames
#
FIFF.FIFFV_MNE_COORD_TUFTS_EEG = 300  # For Tufts EEG data
FIFF.FIFFV_MNE_COORD_CTF_DEVICE = 1001  # CTF device coordinates
FIFF.FIFFV_MNE_COORD_CTF_HEAD = 1004  # CTF head coordinates
FIFF.FIFFV_MNE_COORD_DIGITIZER = (
    FIFF.FIFFV_COORD_ISOTRAK
)  # Original (Polhemus) digitizer coordinates
FIFF.FIFFV_MNE_COORD_SURFACE_RAS = FIFF.FIFFV_COORD_MRI  # The surface RAS coordinates
FIFF.FIFFV_MNE_COORD_MRI_VOXEL = 2001  # The MRI voxel coordinates
FIFF.FIFFV_MNE_COORD_RAS = 2002  # Surface RAS coordinates with non-zero origin
FIFF.FIFFV_MNE_COORD_MNI_TAL = 2003  # MNI Talairach coordinates
FIFF.FIFFV_MNE_COORD_FS_TAL_GTZ = 2004  # FreeSurfer Talairach coordinates (MNI z > 0)
FIFF.FIFFV_MNE_COORD_FS_TAL_LTZ = 2005  # FreeSurfer Talairach coordinates (MNI z < 0)
FIFF.FIFFV_MNE_COORD_FS_TAL = 2006  # FreeSurfer Talairach coordinates
#
# 4D and KIT use the same head coordinate system definition as CTF
#
FIFF.FIFFV_MNE_COORD_4D_HEAD = FIFF.FIFFV_MNE_COORD_CTF_HEAD
FIFF.FIFFV_MNE_COORD_KIT_HEAD = FIFF.FIFFV_MNE_COORD_CTF_HEAD

#
#   FWD Types
#

FWD = BunchConstNamed()

FWD.COIL_UNKNOWN = 0
FWD.COILC_UNKNOWN = 0
FWD.COILC_EEG = 1000
FWD.COILC_MAG = 1
FWD.COILC_AXIAL_GRAD = 2
FWD.COILC_PLANAR_GRAD = 3
FWD.COILC_AXIAL_GRAD2 = 4

FWD.COIL_ACCURACY_POINT = 0
FWD.COIL_ACCURACY_NORMAL = 1
FWD.COIL_ACCURACY_ACCURATE = 2

FWD.BEM_IP_APPROACH_LIMIT = 0.1

FWD.BEM_LIN_FIELD_SIMPLE = 1
FWD.BEM_LIN_FIELD_FERGUSON = 2
FWD.BEM_LIN_FIELD_URANKAR = 3

#
#   Data types
#
FIFF.FIFFT_VOID = 0
FIFF.FIFFT_BYTE = 1
FIFF.FIFFT_SHORT = 2
FIFF.FIFFT_INT = 3
FIFF.FIFFT_FLOAT = 4
FIFF.FIFFT_DOUBLE = 5
FIFF.FIFFT_JULIAN = 6
FIFF.FIFFT_USHORT = 7
FIFF.FIFFT_UINT = 8
FIFF.FIFFT_ULONG = 9
FIFF.FIFFT_STRING = 10
FIFF.FIFFT_LONG = 11
FIFF.FIFFT_DAU_PACK13 = 13
FIFF.FIFFT_DAU_PACK14 = 14
FIFF.FIFFT_DAU_PACK16 = 16
FIFF.FIFFT_COMPLEX_FLOAT = 20
FIFF.FIFFT_COMPLEX_DOUBLE = 21
FIFF.FIFFT_OLD_PACK = 23
FIFF.FIFFT_CH_INFO_STRUCT = 30
FIFF.FIFFT_ID_STRUCT = 31
FIFF.FIFFT_DIR_ENTRY_STRUCT = 32
FIFF.FIFFT_DIG_POINT_STRUCT = 33
FIFF.FIFFT_CH_POS_STRUCT = 34
FIFF.FIFFT_COORD_TRANS_STRUCT = 35
FIFF.FIFFT_DIG_STRING_STRUCT = 36
FIFF.FIFFT_STREAM_SEGMENT_STRUCT = 37
FIFF.FIFFT_MATRIX = 0x40000000  # 1073741824, 1 << 30
FIFF.FIFFT_SPARSE_CCS_MATRIX = 0x00100000  # 1048576
FIFF.FIFFT_SPARSE_RCS_MATRIX = 0x00200000  # 2097152

#
# Units of measurement
#
FIFF.FIFF_UNIT_NONE = -1
#
# SI base units
#
FIFF.FIFF_UNIT_UNITLESS = 0
FIFF.FIFF_UNIT_M = 1  # meter
FIFF.FIFF_UNIT_KG = 2  # kilogram
FIFF.FIFF_UNIT_SEC = 3  # second
FIFF.FIFF_UNIT_A = 4  # ampere
FIFF.FIFF_UNIT_K = 5  # Kelvin
FIFF.FIFF_UNIT_MOL = 6  # mole
#
# SI Supplementary units
#
FIFF.FIFF_UNIT_RAD = 7  # radian
FIFF.FIFF_UNIT_SR = 8  # steradian
#
# SI base candela
#
FIFF.FIFF_UNIT_CD = 9  # candela
#
# SI derived units
#
FIFF.FIFF_UNIT_MOL_M3 = 10  # mol/m^3
FIFF.FIFF_UNIT_HZ = 101  # hertz
FIFF.FIFF_UNIT_N = 102  # Newton
FIFF.FIFF_UNIT_PA = 103  # pascal
FIFF.FIFF_UNIT_J = 104  # joule
FIFF.FIFF_UNIT_W = 105  # watt
FIFF.FIFF_UNIT_C = 106  # coulomb
FIFF.FIFF_UNIT_V = 107  # volt
FIFF.FIFF_UNIT_F = 108  # farad
FIFF.FIFF_UNIT_OHM = 109  # ohm
FIFF.FIFF_UNIT_S = 110  # Siemens (same as Moh, what fiff-constants calls it)
FIFF.FIFF_UNIT_WB = 111  # weber
FIFF.FIFF_UNIT_T = 112  # tesla
FIFF.FIFF_UNIT_H = 113  # Henry
FIFF.FIFF_UNIT_CEL = 114  # celsius
FIFF.FIFF_UNIT_LM = 115  # lumen
FIFF.FIFF_UNIT_LX = 116  # lux
FIFF.FIFF_UNIT_V_M2 = 117  # V/m^2
#
# Others we need
#
FIFF.FIFF_UNIT_T_M = 201  # T/m
FIFF.FIFF_UNIT_AM = 202  # Am
FIFF.FIFF_UNIT_AM_M2 = 203  # Am/m^2
FIFF.FIFF_UNIT_AM_M3 = 204  # Am/m^3

FIFF.FIFF_UNIT_PX = 210  # Pixel
_ch_unit_named = {
    key: key
    for key in (
        FIFF.FIFF_UNIT_NONE,
        FIFF.FIFF_UNIT_UNITLESS,
        FIFF.FIFF_UNIT_M,
        FIFF.FIFF_UNIT_KG,
        FIFF.FIFF_UNIT_SEC,
        FIFF.FIFF_UNIT_A,
        FIFF.FIFF_UNIT_K,
        FIFF.FIFF_UNIT_MOL,
        FIFF.FIFF_UNIT_RAD,
        FIFF.FIFF_UNIT_SR,
        FIFF.FIFF_UNIT_CD,
        FIFF.FIFF_UNIT_MOL_M3,
        FIFF.FIFF_UNIT_HZ,
        FIFF.FIFF_UNIT_N,
        FIFF.FIFF_UNIT_PA,
        FIFF.FIFF_UNIT_J,
        FIFF.FIFF_UNIT_W,
        FIFF.FIFF_UNIT_C,
        FIFF.FIFF_UNIT_V,
        FIFF.FIFF_UNIT_F,
        FIFF.FIFF_UNIT_OHM,
        FIFF.FIFF_UNIT_S,
        FIFF.FIFF_UNIT_WB,
        FIFF.FIFF_UNIT_T,
        FIFF.FIFF_UNIT_H,
        FIFF.FIFF_UNIT_CEL,
        FIFF.FIFF_UNIT_LM,
        FIFF.FIFF_UNIT_LX,
        FIFF.FIFF_UNIT_V_M2,
        FIFF.FIFF_UNIT_T_M,
        FIFF.FIFF_UNIT_AM,
        FIFF.FIFF_UNIT_AM_M2,
        FIFF.FIFF_UNIT_AM_M3,
        FIFF.FIFF_UNIT_PX,
    )
}
#
# Multipliers
#
FIFF.FIFF_UNITM_E = 18
FIFF.FIFF_UNITM_PET = 15
FIFF.FIFF_UNITM_T = 12
FIFF.FIFF_UNITM_GIG = 9
FIFF.FIFF_UNITM_MEG = 6
FIFF.FIFF_UNITM_K = 3
FIFF.FIFF_UNITM_H = 2
FIFF.FIFF_UNITM_DA = 1
FIFF.FIFF_UNITM_NONE = 0
FIFF.FIFF_UNITM_D = -1
FIFF.FIFF_UNITM_C = -2
FIFF.FIFF_UNITM_M = -3
FIFF.FIFF_UNITM_MU = -6
FIFF.FIFF_UNITM_N = -9
FIFF.FIFF_UNITM_P = -12
FIFF.FIFF_UNITM_F = -15
FIFF.FIFF_UNITM_A = -18
_ch_unit_mul_named = {
    key: key
    for key in (
        FIFF.FIFF_UNITM_E,
        FIFF.FIFF_UNITM_PET,
        FIFF.FIFF_UNITM_T,
        FIFF.FIFF_UNITM_GIG,
        FIFF.FIFF_UNITM_MEG,
        FIFF.FIFF_UNITM_K,
        FIFF.FIFF_UNITM_H,
        FIFF.FIFF_UNITM_DA,
        FIFF.FIFF_UNITM_NONE,
        FIFF.FIFF_UNITM_D,
        FIFF.FIFF_UNITM_C,
        FIFF.FIFF_UNITM_M,
        FIFF.FIFF_UNITM_MU,
        FIFF.FIFF_UNITM_N,
        FIFF.FIFF_UNITM_P,
        FIFF.FIFF_UNITM_F,
        FIFF.FIFF_UNITM_A,
    )
}

#
# Coil types
#
FIFF.FIFFV_COIL_NONE = 0  # The location info contains no data
FIFF.FIFFV_COIL_EEG = 1  # EEG electrode position in r0
FIFF.FIFFV_COIL_NM_122 = 2  # Neuromag 122 coils
FIFF.FIFFV_COIL_NM_24 = 3  # Old 24 channel system in HUT
FIFF.FIFFV_COIL_NM_MCG_AXIAL = 4  # The axial devices in the HUCS MCG system
FIFF.FIFFV_COIL_EEG_BIPOLAR = 5  # Bipolar EEG lead
FIFF.FIFFV_COIL_EEG_CSD = 6  # CSD-transformed EEG lead

FIFF.FIFFV_COIL_DIPOLE = 200  # Time-varying dipole definition
# The coil info contains dipole location (r0) and
# direction (ex)
FIFF.FIFFV_COIL_FNIRS_HBO = 300  # fNIRS oxyhemoglobin
FIFF.FIFFV_COIL_FNIRS_HBR = 301  # fNIRS deoxyhemoglobin
FIFF.FIFFV_COIL_FNIRS_CW_AMPLITUDE = 302  # fNIRS continuous wave amplitude
FIFF.FIFFV_COIL_FNIRS_OD = 303  # fNIRS optical density
FIFF.FIFFV_COIL_FNIRS_FD_AC_AMPLITUDE = 304  # fNIRS frequency domain AC amplitude
FIFF.FIFFV_COIL_FNIRS_FD_PHASE = 305  # fNIRS frequency domain phase
FIFF.FIFFV_COIL_FNIRS_RAW = FIFF.FIFFV_COIL_FNIRS_CW_AMPLITUDE  # old alias
FIFF.FIFFV_COIL_FNIRS_TD_GATED_AMPLITUDE = 306  # fNIRS time-domain gated amplitude
FIFF.FIFFV_COIL_FNIRS_TD_MOMENTS_AMPLITUDE = 307  # fNIRS time-domain moments amplitude

FIFF.FIFFV_COIL_EYETRACK_POS = 400  # Eye-tracking gaze position
FIFF.FIFFV_COIL_EYETRACK_PUPIL = 401  # Eye-tracking pupil size

FIFF.FIFFV_COIL_MCG_42 = 1000  # For testing the MCG software

FIFF.FIFFV_COIL_POINT_MAGNETOMETER = 2000  # Simple point magnetometer
FIFF.FIFFV_COIL_AXIAL_GRAD_5CM = 2001  # Generic axial gradiometer

FIFF.FIFFV_COIL_VV_PLANAR_W = 3011  # VV prototype wirewound planar sensor
FIFF.FIFFV_COIL_VV_PLANAR_T1 = 3012  # Vectorview SQ20483N planar gradiometer
FIFF.FIFFV_COIL_VV_PLANAR_T2 = 3013  # Vectorview SQ20483N-A planar gradiometer
FIFF.FIFFV_COIL_VV_PLANAR_T3 = 3014  # Vectorview SQ20950N planar gradiometer
FIFF.FIFFV_COIL_VV_PLANAR_T4 = 3015  # Vectorview planar gradiometer (MEG-MRI)
FIFF.FIFFV_COIL_VV_MAG_W = 3021  # VV prototype wirewound magnetometer
FIFF.FIFFV_COIL_VV_MAG_T1 = 3022  # Vectorview SQ20483N magnetometer
FIFF.FIFFV_COIL_VV_MAG_T2 = 3023  # Vectorview SQ20483-A magnetometer
FIFF.FIFFV_COIL_VV_MAG_T3 = 3024  # Vectorview SQ20950N magnetometer
FIFF.FIFFV_COIL_VV_MAG_T4 = 3025  # Vectorview magnetometer (MEG-MRI)

FIFF.FIFFV_COIL_MAGNES_MAG = 4001  # Magnes WH magnetometer
FIFF.FIFFV_COIL_MAGNES_GRAD = 4002  # Magnes WH gradiometer
#
# Magnes reference sensors
#
FIFF.FIFFV_COIL_MAGNES_REF_MAG = 4003
FIFF.FIFFV_COIL_MAGNES_REF_GRAD = 4004
FIFF.FIFFV_COIL_MAGNES_OFFDIAG_REF_GRAD = 4005
FIFF.FIFFV_COIL_MAGNES_R_MAG = FIFF.FIFFV_COIL_MAGNES_REF_MAG
FIFF.FIFFV_COIL_MAGNES_R_GRAD = FIFF.FIFFV_COIL_MAGNES_REF_GRAD
FIFF.FIFFV_COIL_MAGNES_R_GRAD_OFF = FIFF.FIFFV_COIL_MAGNES_OFFDIAG_REF_GRAD

#
# CTF coil and channel types
#
FIFF.FIFFV_COIL_CTF_GRAD = 5001
FIFF.FIFFV_COIL_CTF_REF_MAG = 5002
FIFF.FIFFV_COIL_CTF_REF_GRAD = 5003
FIFF.FIFFV_COIL_CTF_OFFDIAG_REF_GRAD = 5004
#
# KIT system coil types
#
FIFF.FIFFV_COIL_KIT_GRAD = 6001
FIFF.FIFFV_COIL_KIT_REF_MAG = 6002
#
# BabySQUID sensors
#
FIFF.FIFFV_COIL_BABY_GRAD = 7001
#
# BabyMEG sensors
#
FIFF.FIFFV_COIL_BABY_MAG = 7002
FIFF.FIFFV_COIL_BABY_REF_MAG = 7003
FIFF.FIFFV_COIL_BABY_REF_MAG2 = 7004
#
# Artemis123 sensors
#
FIFF.FIFFV_COIL_ARTEMIS123_GRAD = 7501
FIFF.FIFFV_COIL_ARTEMIS123_REF_MAG = 7502
FIFF.FIFFV_COIL_ARTEMIS123_REF_GRAD = 7503
#
# QuSpin sensors
#
FIFF.FIFFV_COIL_QUSPIN_ZFOPM_MAG = 8001
FIFF.FIFFV_COIL_QUSPIN_ZFOPM_MAG2 = 8002
#
# FieldLine sensors
#
FIFF.FIFFV_COIL_FIELDLINE_OPM_MAG_GEN1 = 8101
#
# Kernel sensors
#
FIFF.FIFFV_COIL_KERNEL_OPM_MAG_GEN1 = 8201
#
# KRISS sensors
#
FIFF.FIFFV_COIL_KRISS_GRAD = 9001
#
# Compumedics adult/pediatric gradiometer
#
FIFF.FIFFV_COIL_COMPUMEDICS_ADULT_GRAD = 9101
FIFF.FIFFV_COIL_COMPUMEDICS_PEDIATRIC_GRAD = 9102
_ch_coil_type_named = {
    key: key
    for key in (
        FIFF.FIFFV_COIL_NONE,
        FIFF.FIFFV_COIL_EEG,
        FIFF.FIFFV_COIL_NM_122,
        FIFF.FIFFV_COIL_NM_24,
        FIFF.FIFFV_COIL_NM_MCG_AXIAL,
        FIFF.FIFFV_COIL_EEG_BIPOLAR,
        FIFF.FIFFV_COIL_EEG_CSD,
        FIFF.FIFFV_COIL_DIPOLE,
        FIFF.FIFFV_COIL_FNIRS_HBO,
        FIFF.FIFFV_COIL_FNIRS_HBR,
        FIFF.FIFFV_COIL_FNIRS_RAW,
        FIFF.FIFFV_COIL_FNIRS_OD,
        FIFF.FIFFV_COIL_FNIRS_FD_AC_AMPLITUDE,
        FIFF.FIFFV_COIL_FNIRS_FD_PHASE,
        FIFF.FIFFV_COIL_FNIRS_TD_GATED_AMPLITUDE,
        FIFF.FIFFV_COIL_FNIRS_TD_MOMENTS_AMPLITUDE,
        FIFF.FIFFV_COIL_MCG_42,
        FIFF.FIFFV_COIL_EYETRACK_POS,
        FIFF.FIFFV_COIL_EYETRACK_PUPIL,
        FIFF.FIFFV_COIL_POINT_MAGNETOMETER,
        FIFF.FIFFV_COIL_AXIAL_GRAD_5CM,
        FIFF.FIFFV_COIL_VV_PLANAR_W,
        FIFF.FIFFV_COIL_VV_PLANAR_T1,
        FIFF.FIFFV_COIL_VV_PLANAR_T2,
        FIFF.FIFFV_COIL_VV_PLANAR_T3,
        FIFF.FIFFV_COIL_VV_PLANAR_T4,
        FIFF.FIFFV_COIL_VV_MAG_W,
        FIFF.FIFFV_COIL_VV_MAG_T1,
        FIFF.FIFFV_COIL_VV_MAG_T2,
        FIFF.FIFFV_COIL_VV_MAG_T3,
        FIFF.FIFFV_COIL_VV_MAG_T4,
        FIFF.FIFFV_COIL_MAGNES_MAG,
        FIFF.FIFFV_COIL_MAGNES_GRAD,
        FIFF.FIFFV_COIL_MAGNES_REF_MAG,
        FIFF.FIFFV_COIL_MAGNES_REF_GRAD,
        FIFF.FIFFV_COIL_MAGNES_OFFDIAG_REF_GRAD,
        FIFF.FIFFV_COIL_CTF_GRAD,
        FIFF.FIFFV_COIL_CTF_REF_MAG,
        FIFF.FIFFV_COIL_CTF_REF_GRAD,
        FIFF.FIFFV_COIL_CTF_OFFDIAG_REF_GRAD,
        FIFF.FIFFV_COIL_KIT_GRAD,
        FIFF.FIFFV_COIL_KIT_REF_MAG,
        FIFF.FIFFV_COIL_BABY_GRAD,
        FIFF.FIFFV_COIL_BABY_MAG,
        FIFF.FIFFV_COIL_BABY_REF_MAG,
        FIFF.FIFFV_COIL_BABY_REF_MAG2,
        FIFF.FIFFV_COIL_ARTEMIS123_GRAD,
        FIFF.FIFFV_COIL_ARTEMIS123_REF_MAG,
        FIFF.FIFFV_COIL_ARTEMIS123_REF_GRAD,
        FIFF.FIFFV_COIL_QUSPIN_ZFOPM_MAG,
        FIFF.FIFFV_COIL_QUSPIN_ZFOPM_MAG2,
        FIFF.FIFFV_COIL_FIELDLINE_OPM_MAG_GEN1,
        FIFF.FIFFV_COIL_KERNEL_OPM_MAG_GEN1,
        FIFF.FIFFV_COIL_KRISS_GRAD,
        FIFF.FIFFV_COIL_COMPUMEDICS_ADULT_GRAD,
        FIFF.FIFFV_COIL_COMPUMEDICS_PEDIATRIC_GRAD,
    )
}

# MNE RealTime
FIFF.FIFF_MNE_RT_COMMAND = 3700  # realtime command
FIFF.FIFF_MNE_RT_CLIENT_ID = 3701  # realtime client

# MNE epochs bookkeeping
FIFF.FIFF_MNE_EPOCHS_SELECTION = 3800  # the epochs selection
FIFF.FIFF_MNE_EPOCHS_DROP_LOG = 3801  # the drop log
FIFF.FIFF_MNE_EPOCHS_REJECT_FLAT = 3802  # rejection and flat params
FIFF.FIFF_MNE_EPOCHS_RAW_SFREQ = 3803  # original raw sfreq

# MNE annotations
FIFF.FIFFB_MNE_ANNOTATIONS = 3810  # annotations block

# MNE Metadata Dataframes
FIFF.FIFFB_MNE_METADATA = 3811  # metadata dataframes block

# Table to match unrecognized channel location names to their known aliases
CHANNEL_LOC_ALIASES = {
    # this set of aliases are published in doi:10.1097/WNP.0000000000000316 and
    # doi:10.1016/S1388-2457(00)00527-7.
    "Cb1": "POO7",
    "Cb2": "POO8",
    "CB1": "POO7",
    "CB2": "POO8",
    "T1": "T9",
    "T2": "T10",
    "T3": "T7",
    "T4": "T8",
    "T5": "T9",
    "T6": "T10",
    "M1": "TP9",
    "M2": "TP10",
    # EGI ref chan is named VREF/Vertex Ref.
    # In the standard montages for EGI, the ref is named Cz
    "VREF": "Cz",
    "Vertex Reference": "Cz"
    # add a comment here (with doi of a published source) above any new
    # aliases, as they are added
}



"""File downloading functions."""

# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

import os


def _url_to_local_path(url, path):
    """Mirror a url path in a local destination (keeping folder structure)."""
    from urllib import parse, request

    destination = parse.urlparse(url).path
    # First char should be '/', and it needs to be discarded
    if len(destination) < 2 or destination[0] != "/":
        raise ValueError("Invalid URL")
    destination = os.path.join(path, request.url2pathname(destination)[1:])
    return destination



# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

import warnings
from functools import partial

import numpy as np
from scipy.signal import spectrogram

from ..parallel import parallel_func
from ..utils import _check_option, _ensure_int, logger, verbose
from ..utils.numerics import _mask_to_onsets_offsets


# adapted from SciPy
# https://github.com/scipy/scipy/blob/f71e7fad717801c4476312fe1e23f2dfbb4c9d7f/scipy/signal/_spectral_py.py#L2019  # noqa: E501
def _median_biases(n):
    # Compute the biases for 0 to max(n, 1) terms included in a median calc
    biases = np.ones(n + 1)
    # The original SciPy code is:
    #
    # def _median_bias(n):
    #     ii_2 = 2 * np.arange(1., (n - 1) // 2 + 1)
    #     return 1 + np.sum(1. / (ii_2 + 1) - 1. / ii_2)
    #
    # This is a sum over (n-1)//2 terms.
    # The ii_2 terms here for different n are:
    #
    # n=0: []  # 0 terms
    # n=1: []  # 0 terms
    # n=2: []  # 0 terms
    # n=3: [2]  # 1 term
    # n=4: [2]  # 1 term
    # n=5: [2, 4]  # 2 terms
    # n=6: [2, 4]  # 2 terms
    # ...
    #
    # We can get the terms for 0 through n using a cumulative summation and
    # indexing:
    if n >= 3:
        ii_2 = 2 * np.arange(1, (n - 1) // 2 + 1)
        sums = 1 + np.cumsum(1.0 / (ii_2 + 1) - 1.0 / ii_2)
        idx = np.arange(2, n) // 2 - 1
        biases[3:] = sums[idx]
    return biases


def _decomp_aggregate_mask(epoch, func, average, freq_sl):
    _, _, spect = func(epoch)
    spect = spect[..., freq_sl, :]
    # Do the averaging here (per epoch) to save memory
    if average == "mean":
        spect = np.nanmean(spect, axis=-1)
    elif average == "median":
        biases = _median_biases(spect.shape[-1])
        idx = (~np.isnan(spect)).sum(-1)
        spect = np.nanmedian(spect, axis=-1) / biases[idx]
    return spect


def _spect_func(epoch, func, freq_sl, average, *, output="power"):
    """Aux function."""
    # Decide if we should split this to save memory or not, since doing
    # multiple calls will incur some performance overhead. Eventually we might
    # want to write (really, go back to) our own spectrogram implementation
    # that, if possible, averages after each transform, but this will incur
    # a lot of overhead because of the many Python calls required.
    kwargs = dict(func=func, average=average, freq_sl=freq_sl)
    if epoch.nbytes > 10e6:
        spect = np.apply_along_axis(_decomp_aggregate_mask, -1, epoch, **kwargs)
    else:
        spect = _decomp_aggregate_mask(epoch, **kwargs)
    return spect


def _check_nfft(n, n_fft, n_per_seg, n_overlap):
    """Ensure n_fft, n_per_seg and n_overlap make sense."""
    if n_per_seg is None and n_fft > n:
        raise ValueError(
            "If n_per_seg is None n_fft is not allowed to be > "
            "n_times. If you want zero-padding, you have to set "
            f"n_per_seg to relevant length. Got n_fft of {n_fft} while"
            f" signal length is {n}."
        )
    n_per_seg = n_fft if n_per_seg is None or n_per_seg > n_fft else n_per_seg
    n_per_seg = n if n_per_seg > n else n_per_seg
    if n_overlap >= n_per_seg:
        raise ValueError(
            "n_overlap cannot be greater than n_per_seg (or n_fft). Got n_overlap "
            f"of {n_overlap} while n_per_seg is {n_per_seg}."
        )
    return n_fft, n_per_seg, n_overlap


@verbose
def psd_array_welch(
    x,
    sfreq,
    fmin=0,
    fmax=np.inf,
    n_fft=256,
    n_overlap=0,
    n_per_seg=None,
    n_jobs=None,
    average="mean",
    window="hamming",
    remove_dc=True,
    *,
    output="power",
    verbose=None,
):
    """Compute power spectral density (PSD) using Welch's method.

    Welch's method is described in :footcite:t:`Welch1967`.

    Parameters
    ----------
    x : array, shape=(..., n_times)
        The data to compute PSD from.
    sfreq : float
        The sampling frequency.
    fmin : float
        The lower frequency of interest.
    fmax : float
        The upper frequency of interest.
    n_fft : int
        The length of FFT used, must be ``>= n_per_seg`` (default: 256).
        The segments will be zero-padded if ``n_fft > n_per_seg``.
    n_overlap : int
        The number of points of overlap between segments. Will be adjusted
        to be <= n_per_seg. The default value is 0.
    n_per_seg : int | None
        Length of each Welch segment (windowed with a Hamming window). Defaults
        to None, which sets n_per_seg equal to n_fft.
    %(n_jobs)s
    %(average_psd)s

        .. versionadded:: 0.19.0
    %(window_psd)s

        .. versionadded:: 0.22.0
    %(remove_dc)s

    output : str
        The format of the returned ``psds`` array, ``'complex'`` or
        ``'power'``:

        * ``'power'`` : the power spectral density is returned.
        * ``'complex'`` : the complex fourier coefficients are returned per
          window.

        .. versionadded:: 1.4.0
    %(verbose)s

    Returns
    -------
    psds : ndarray, shape (..., n_freqs) or (..., n_freqs, n_segments)
        The power spectral densities. If ``average='mean`` or
        ``average='median'``, the returned array will have the same shape
        as the input data plus an additional frequency dimension.
        If ``average=None``, the returned array will have the same shape as
        the input data plus two additional dimensions corresponding to
        frequencies and the unaggregated segments, respectively.
    freqs : ndarray, shape (n_freqs,)
        The frequencies.

    Notes
    -----
    .. versionadded:: 0.14.0

    References
    ----------
    .. footbibliography::
    """
    _check_option("average", average, (None, False, "mean", "median"))
    _check_option("output", output, ("power", "complex"))
    detrend = "constant" if remove_dc else False
    mode = "complex" if output == "complex" else "psd"
    n_fft = _ensure_int(n_fft, "n_fft")
    n_overlap = _ensure_int(n_overlap, "n_overlap")
    if n_per_seg is not None:
        n_per_seg = _ensure_int(n_per_seg, "n_per_seg")
    if average is False:
        average = None

    dshape = x.shape[:-1]
    n_times = x.shape[-1]
    x = x.reshape(-1, n_times)

    # Prep the PSD
    n_fft, n_per_seg, n_overlap = _check_nfft(n_times, n_fft, n_per_seg, n_overlap)
    win_size = n_fft / float(sfreq)
    logger.info(f"Effective window size : {win_size:0.3f} (s)")
    freqs = np.arange(n_fft // 2 + 1, dtype=float) * (sfreq / n_fft)
    freq_mask = (freqs >= fmin) & (freqs <= fmax)
    if not freq_mask.any():
        raise ValueError(f"No frequencies found between fmin={fmin} and fmax={fmax}")
    freq_sl = slice(*(np.where(freq_mask)[0][[0, -1]] + [0, 1]))
    del freq_mask
    freqs = freqs[freq_sl]

    # Parallelize across first N-1 dimensions
    logger.debug(
        f"Spectogram using {n_fft}-point FFT on {n_per_seg} samples with "
        f"{n_overlap} overlap and {window} window"
    )

    parallel, my_spect_func, n_jobs = parallel_func(_spect_func, n_jobs=n_jobs)
    _func = partial(
        spectrogram,
        detrend=detrend,
        noverlap=n_overlap,
        nperseg=n_per_seg,
        nfft=n_fft,
        fs=sfreq,
        window=window,
        mode=mode,
    )
    if np.any(np.isnan(x)):
        good_mask = ~np.isnan(x)
        # NaNs originate from annot, so must match for all channels. Note that we CANNOT
        # use np.testing.assert_allclose() here; it is strict about shapes/broadcasting
        assert np.allclose(good_mask, good_mask[[0]], equal_nan=True)
        t_onsets, t_offsets = _mask_to_onsets_offsets(good_mask[0])
        x_splits = [x[..., t_ons:t_off] for t_ons, t_off in zip(t_onsets, t_offsets)]
        # weights reflect the number of samples used from each span. For spans longer
        # than `n_per_seg`, trailing samples may be discarded. For spans shorter than
        # `n_per_seg`, the wrapped function (`scipy.signal.spectrogram`) automatically
        # reduces `n_per_seg` to match the span length (with a warning).
        step = n_per_seg - n_overlap
        span_lengths = [span.shape[-1] for span in x_splits]
        weights = [
            w if w < n_per_seg else w - ((w - n_overlap) % step) for w in span_lengths
        ]
        agg_func = partial(np.average, weights=weights)
        if n_jobs > 1:
            logger.info(
                f"Data split into {len(x_splits)} (probably unequal) chunks due to "
                '"bad_*" annotations. Parallelization may be sub-optimal.'
            )
        if (np.array(span_lengths) < n_per_seg).any():
            logger.info(
                "At least one good data span is shorter than n_per_seg, and will be "
                "analyzed with a shorter window than the rest of the file."
            )

        def func(*args, **kwargs):
            # swallow SciPy warnings caused by short good data spans
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    action="ignore",
                    module="scipy",
                    category=UserWarning,
                    message=r"nperseg = \d+ is greater than input length",
                )
                return _func(*args, **kwargs)

    else:
        x_splits = [arr for arr in np.array_split(x, n_jobs) if arr.size != 0]
        agg_func = np.concatenate
        func = _func
    f_spect = parallel(
        my_spect_func(d, func=func, freq_sl=freq_sl, average=average, output=output)
        for d in x_splits
    )
    psds = agg_func(f_spect, axis=0)
    shape = dshape + (len(freqs),)
    if average is None:
        shape = shape + (-1,)
    psds.shape = shape
    return psds, freqs



"""Compatibility fixes for older versions of libraries.

If you add content to this file, please give the version of the package
at which the fix is no longer needed.

# originally copied from scikit-learn

"""

# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

# NOTE:
# Imports for SciPy submodules need to stay nested in this module
# because this module is imported many places (but not always used)!

import inspect
import operator as operator_module
import os
import warnings
from io import StringIO
from math import log
from pprint import pprint

import numpy as np

###############################################################################
# distutils

# distutils has been deprecated since Python 3.10 and was removed
# from the standard library with the release of Python 3.12.


def _compare_version(version_a, operator, version_b):
    """Compare two version strings via a user-specified operator.

    Parameters
    ----------
    version_a : str
        First version string.
    operator : '==' | '>' | '<' | '>=' | '<='
        Operator to compare ``version_a`` and ``version_b`` in the form of
        ``version_a operator version_b``.
    version_b : str
        Second version string.

    Returns
    -------
    bool
        The result of the version comparison.
    """
    from packaging.version import parse

    mapping = {"<": "lt", "<=": "le", "==": "eq", "!=": "ne", ">=": "ge", ">": "gt"}
    with warnings.catch_warnings(record=True):
        warnings.simplefilter("ignore")
        ver_a = parse(version_a)
        ver_b = parse(version_b)
        return getattr(operator_module, mapping[operator])(ver_a, ver_b)


###############################################################################
# Misc


def _median_complex(data, axis):
    """Compute marginal median on complex data safely.

    Can be removed when numpy introduces a fix.
    See: https://github.com/scipy/scipy/pull/12676/.
    """
    # np.median must be passed real arrays for the desired result
    if np.iscomplexobj(data):
        data = np.median(np.real(data), axis=axis) + 1j * np.median(
            np.imag(data), axis=axis
        )
    else:
        data = np.median(data, axis=axis)
    return data


def _safe_svd(A, **kwargs):
    """Get around the SVD did not converge error of death."""
    # Intel has a bug with their GESVD driver:
    #     https://software.intel.com/en-us/forums/intel-distribution-for-python/topic/628049  # noqa: E501
    # For SciPy 0.18 and up, we can work around it by using
    # lapack_driver='gesvd' instead.
    from scipy import linalg

    if kwargs.get("overwrite_a", False):
        raise ValueError("Cannot set overwrite_a=True with this function")
    try:
        return linalg.svd(A, **kwargs)
    except np.linalg.LinAlgError as exp:
        from .utils import warn

        warn(f"SVD error ({exp}), attempting to use GESVD instead of GESDD")
        return linalg.svd(A, lapack_driver="gesvd", **kwargs)


def _csc_array_cast(x):
    from scipy.sparse import csc_array

    return csc_array(x)


# Can be replaced with sparse.eye_array once we depend on SciPy >= 1.12
def _eye_array(n, *, format="csr"):  # noqa: A002
    from scipy import sparse

    return sparse.dia_array((np.ones(n), 0), shape=(n, n)).asformat(format)


###############################################################################
# NumPy Generator (NumPy 1.17)


def rng_uniform(rng):
    """Get the uniform/randint from the rng."""
    # prefer Generator.integers, fall back to RandomState.randint
    return getattr(rng, "integers", getattr(rng, "randint", None))


###############################################################################
# Misc utilities


# get_fdata() requires knowing the dtype ahead of time, so let's triage on our
# own instead
def _get_img_fdata(img):
    data = np.asanyarray(img.dataobj)
    dtype = np.complex128 if np.iscomplexobj(data) else np.float64
    return data.astype(dtype)


##############################################################################
# adapted from scikit-learn


_DEFAULT_TAGS = {
    "array_api_support": False,
    "non_deterministic": False,
    "requires_positive_X": False,
    "requires_positive_y": False,
    "X_types": ["2darray"],
    "poor_score": False,
    "no_validation": False,
    "multioutput": False,
    "allow_nan": False,
    "stateless": False,
    "multilabel": False,
    "_skip_test": False,
    "_xfail_checks": False,
    "multioutput_only": False,
    "binary_only": False,
    "requires_fit": True,
    "preserves_dtype": [np.float64],
    "requires_y": False,
    "pairwise": False,
}


class BaseEstimator:
    """Base class for all estimators in scikit-learn.

    Notes
    -----
    All estimators should specify all the parameters that can be set
    at the class level in their ``__init__`` as explicit keyword
    arguments (no ``*args`` or ``**kwargs``).
    """

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator."""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, "deprecated_original", cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [
            p
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError(
                    "scikit-learn estimators should always "
                    "specify their parameters in the signature"
                    " of their __init__ (no varargs)."
                    f" {cls} with constructor {init_signature} doesn't "
                    " follow this convention."
                )
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            # We need deprecation warnings to always be on in order to
            # catch deprecated param values.
            # This is set in utils/__init__.py but it gets overwritten
            # when running under python3 somehow.
            warnings.simplefilter("always", DeprecationWarning)
            try:
                with warnings.catch_warnings(record=True) as w:
                    value = getattr(self, key, None)
                if len(w) and w[0].category is DeprecationWarning:
                    # if the parameter is deprecated, don't show it
                    continue
            finally:
                warnings.filters.pop(0)

            # XXX: should we rather test if instance of estimator?
            if deep and hasattr(value, "get_params"):
                deep_items = value.get_params().items()
                out.update((key + "__" + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Parameters
        ----------
        **params : dict
            Parameters.

        Returns
        -------
        inst : instance
            The object.
        """
        if not params:
            # Simple optimisation to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)
        for key, value in params.items():
            split = key.split("__", 1)
            if len(split) > 1:
                # nested objects case
                name, sub_name = split
                if name not in valid_params:
                    raise ValueError(
                        f"Invalid parameter {name} for estimator {self}. "
                        "Check the list of available parameters "
                        "with `estimator.get_params().keys()`."
                    )
                sub_object = valid_params[name]
                sub_object.set_params(**{sub_name: value})
            else:
                # simple objects case
                if key not in valid_params:
                    raise ValueError(
                        f"Invalid parameter {key} for estimator "
                        f"{self.__class__.__name__}. "
                        "Check the list of available parameters "
                        "with `estimator.get_params().keys()`."
                    )
                setattr(self, key, value)
        return self

    def __repr__(self):  # noqa: D105
        params = StringIO()
        pprint(self.get_params(deep=False), params)
        params.seek(0)
        class_name = self.__class__.__name__
        return f"{class_name}({params.read().strip()})"

    # __getstate__ and __setstate__ are omitted because they only contain
    # conditionals that are not satisfied by our objects (e.g.,
    # ``if type(self).__module__.startswith('sklearn.')``.

    def _more_tags(self):
        return _DEFAULT_TAGS

    def _get_tags(self):
        collected_tags = {}
        for base_class in reversed(inspect.getmro(self.__class__)):
            if hasattr(base_class, "_more_tags"):
                # need the if because mixins might not have _more_tags
                # but might do redundant work in estimators
                # (i.e. calling more tags on BaseEstimator multiple times)
                more_tags = base_class._more_tags(self)
                collected_tags.update(more_tags)
        return collected_tags


# newer sklearn deprecates importing from sklearn.metrics.scoring,
# but older sklearn does not expose check_scoring in sklearn.metrics.
def _get_check_scoring():
    try:
        from sklearn.metrics import check_scoring  # noqa
    except ImportError:
        from sklearn.metrics.scorer import check_scoring  # noqa
    return check_scoring


def _check_fit_params(X, fit_params, indices=None):
    """Check and validate the parameters passed during `fit`.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        Data array.

    fit_params : dict
        Dictionary containing the parameters passed at fit.

    indices : array-like of shape (n_samples,), default=None
        Indices to be selected if the parameter has the same size as
        `X`.

    Returns
    -------
    fit_params_validated : dict
        Validated parameters. We ensure that the values support
        indexing.
    """
    try:
        from sklearn.utils.validation import (
            _check_fit_params as _sklearn_check_fit_params,
        )

        return _sklearn_check_fit_params(X, fit_params, indices)
    except ImportError:
        from sklearn.model_selection import _validation

        fit_params_validated = {
            k: _validation._index_param_value(X, v, indices)
            for k, v in fit_params.items()
        }
        return fit_params_validated


###############################################################################
# Copied from sklearn to simplify code paths


def empirical_covariance(X, assume_centered=False):
    """Compute the Maximum likelihood covariance estimator.

    Parameters
    ----------
    X : ndarray, shape (n_samples, n_features)
        Data from which to compute the covariance estimate

    assume_centered : Boolean
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False, data are centered before computation.

    Returns
    -------
    covariance : 2D ndarray, shape (n_features, n_features)
        Empirical covariance (Maximum Likelihood Estimator).
    """
    X = np.asarray(X)
    if X.ndim == 1:
        X = np.reshape(X, (1, -1))

    if X.shape[0] == 1:
        warnings.warn(
            "Only one sample available. You may want to reshape your data array"
        )

    if assume_centered:
        covariance = np.dot(X.T, X) / X.shape[0]
    else:
        covariance = np.cov(X.T, bias=1)

    if covariance.ndim == 0:
        covariance = np.array([[covariance]])
    return covariance


class EmpiricalCovariance(BaseEstimator):
    """Maximum likelihood covariance estimator.

    Read more in the :ref:`User Guide <covariance>`.

    Parameters
    ----------
    store_precision : bool
        Specifies if the estimated precision is stored.

    assume_centered : bool
        If True, data are not centered before computation.
        Useful when working with data whose mean is almost, but not exactly
        zero.
        If False (default), data are centered before computation.

    Attributes
    ----------
    covariance_ : 2D ndarray, shape (n_features, n_features)
        Estimated covariance matrix

    precision_ : 2D ndarray, shape (n_features, n_features)
        Estimated pseudo-inverse matrix.
        (stored only if store_precision is True)
    """

    def __init__(self, store_precision=True, assume_centered=False):
        self.store_precision = store_precision
        self.assume_centered = assume_centered

    def _set_covariance(self, covariance):
        """Save the covariance and precision estimates.

        Storage is done accordingly to `self.store_precision`.
        Precision stored only if invertible.

        Parameters
        ----------
        covariance : 2D ndarray, shape (n_features, n_features)
            Estimated covariance matrix to be stored, and from which precision
            is computed.
        """
        from scipy import linalg

        # covariance = check_array(covariance)
        # set covariance
        self.covariance_ = covariance
        # set precision
        if self.store_precision:
            self.precision_ = linalg.pinvh(covariance)
        else:
            self.precision_ = None

    def get_precision(self):
        """Getter for the precision matrix.

        Returns
        -------
        precision_ : array-like,
            The precision matrix associated to the current covariance object.

        """
        from scipy import linalg

        if self.store_precision:
            precision = self.precision_
        else:
            precision = linalg.pinvh(self.covariance_)
        return precision

    def fit(self, X, y=None):
        """Fit the Maximum Likelihood Estimator covariance model.

        Parameters
        ----------
        X : array-like, shape = [n_samples, n_features]
          Training data, where n_samples is the number of samples and
          n_features is the number of features.
        y : ndarray | None
            Not used, present for API consistency.

        Returns
        -------
        self : object
            Returns self.
        """  # noqa: E501
        # X = check_array(X)
        if self.assume_centered:
            self.location_ = np.zeros(X.shape[1])
        else:
            self.location_ = X.mean(0)
        covariance = empirical_covariance(X, assume_centered=self.assume_centered)
        self._set_covariance(covariance)

        return self

    def score(self, X_test, y=None):
        """Compute the log-likelihood of a Gaussian dataset.

        Uses ``self.covariance_`` as an estimator of its covariance matrix.

        Parameters
        ----------
        X_test : array-like, shape = [n_samples, n_features]
            Test data of which we compute the likelihood, where n_samples is
            the number of samples and n_features is the number of features.
            X_test is assumed to be drawn from the same distribution than
            the data used in fit (including centering).
        y : ndarray | None
            Not used, present for API consistency.

        Returns
        -------
        res : float
            The likelihood of the data set with `self.covariance_` as an
            estimator of its covariance matrix.
        """
        # compute empirical covariance of the test set
        test_cov = empirical_covariance(X_test - self.location_, assume_centered=True)
        # compute log likelihood
        res = log_likelihood(test_cov, self.get_precision())

        return res

    def error_norm(self, comp_cov, norm="frobenius", scaling=True, squared=True):
        """Compute the Mean Squared Error between two covariance estimators.

        Parameters
        ----------
        comp_cov : array-like, shape = [n_features, n_features]
            The covariance to compare with.
        norm : str
            The type of norm used to compute the error. Available error types:
            - 'frobenius' (default): sqrt(tr(A^t.A))
            - 'spectral': sqrt(max(eigenvalues(A^t.A))
            where A is the error ``(comp_cov - self.covariance_)``.
        scaling : bool
            If True (default), the squared error norm is divided by n_features.
            If False, the squared error norm is not rescaled.
        squared : bool
            Whether to compute the squared error norm or the error norm.
            If True (default), the squared error norm is returned.
            If False, the error norm is returned.

        Returns
        -------
        The Mean Squared Error (in the sense of the Frobenius norm) between
        `self` and `comp_cov` covariance estimators.
        """
        from scipy import linalg

        # compute the error
        error = comp_cov - self.covariance_
        # compute the error norm
        if norm == "frobenius":
            squared_norm = np.sum(error**2)
        elif norm == "spectral":
            squared_norm = np.amax(linalg.svdvals(np.dot(error.T, error)))
        else:
            raise NotImplementedError(
                "Only spectral and frobenius norms are implemented"
            )
        # optionally scale the error norm
        if scaling:
            squared_norm = squared_norm / error.shape[0]
        # finally get either the squared norm or the norm
        if squared:
            result = squared_norm
        else:
            result = np.sqrt(squared_norm)

        return result

    def mahalanobis(self, observations):
        """Compute the squared Mahalanobis distances of given observations.

        Parameters
        ----------
        observations : array-like, shape = [n_observations, n_features]
            The observations, the Mahalanobis distances of the which we
            compute. Observations are assumed to be drawn from the same
            distribution than the data used in fit.

        Returns
        -------
        mahalanobis_distance : array, shape = [n_observations,]
            Squared Mahalanobis distances of the observations.
        """
        precision = self.get_precision()
        # compute mahalanobis distances
        centered_obs = observations - self.location_
        mahalanobis_dist = np.sum(np.dot(centered_obs, precision) * centered_obs, 1)

        return mahalanobis_dist


def log_likelihood(emp_cov, precision):
    """Compute the sample mean of the log_likelihood under a covariance model.

    computes the empirical expected log-likelihood (accounting for the
    normalization terms and scaling), allowing for universal comparison (beyond
    this software package)

    Parameters
    ----------
    emp_cov : 2D ndarray (n_features, n_features)
        Maximum Likelihood Estimator of covariance

    precision : 2D ndarray (n_features, n_features)
        The precision matrix of the covariance model to be tested

    Returns
    -------
    sample mean of the log-likelihood
    """
    p = precision.shape[0]
    log_likelihood_ = -np.sum(emp_cov * precision) + _logdet(precision)
    log_likelihood_ -= p * np.log(2 * np.pi)
    log_likelihood_ /= 2.0
    return log_likelihood_


# sklearn uses np.linalg for this, but ours is more robust to zero eigenvalues


def _logdet(A):
    """Compute the log det of a positive semidefinite matrix."""
    from scipy import linalg

    vals = linalg.eigvalsh(A)
    # avoid negative (numerical errors) or zero (semi-definite matrix) values
    tol = vals.max() * vals.size * np.finfo(np.float64).eps
    vals = np.where(vals > tol, vals, tol)
    return np.sum(np.log(vals))


def _infer_dimension_(spectrum, n_samples, n_features):
    """Infer the dimension of a dataset of shape (n_samples, n_features).

    The dataset is described by its spectrum `spectrum`.
    """
    n_spectrum = len(spectrum)
    ll = np.empty(n_spectrum)
    for rank in range(n_spectrum):
        ll[rank] = _assess_dimension_(spectrum, rank, n_samples, n_features)
    return ll.argmax()


def _assess_dimension_(spectrum, rank, n_samples, n_features):
    from scipy.special import gammaln

    if rank > len(spectrum):
        raise ValueError("The tested rank cannot exceed the rank of the dataset")

    pu = -rank * log(2.0)
    for i in range(rank):
        pu += gammaln((n_features - i) / 2.0) - log(np.pi) * (n_features - i) / 2.0

    pl = np.sum(np.log(spectrum[:rank]))
    pl = -pl * n_samples / 2.0

    if rank == n_features:
        pv = 0
        v = 1
    else:
        v = np.sum(spectrum[rank:]) / (n_features - rank)
        pv = -np.log(v) * n_samples * (n_features - rank) / 2.0

    m = n_features * rank - rank * (rank + 1.0) / 2.0
    pp = log(2.0 * np.pi) * (m + rank + 1.0) / 2.0

    pa = 0.0
    spectrum_ = spectrum.copy()
    spectrum_[rank:n_features] = v
    for i in range(rank):
        for j in range(i + 1, len(spectrum)):
            pa += log(
                (spectrum[i] - spectrum[j]) * (1.0 / spectrum_[j] - 1.0 / spectrum_[i])
            ) + log(n_samples)

    ll = pu + pl + pv + pp - pa / 2.0 - rank * log(n_samples) / 2.0

    return ll


def svd_flip(u, v, u_based_decision=True):  # noqa: D103
    if u_based_decision:
        # columns of u, rows of v
        max_abs_cols = np.argmax(np.abs(u), axis=0)
        signs = np.sign(u[max_abs_cols, np.arange(u.shape[1])])
        u *= signs
        v *= signs[:, np.newaxis]
    else:
        # rows of v, columns of u
        max_abs_rows = np.argmax(np.abs(v), axis=1)
        signs = np.sign(v[np.arange(v.shape[0]), max_abs_rows])
        u *= signs
        v *= signs[:, np.newaxis]
    return u, v


def stable_cumsum(arr, axis=None, rtol=1e-05, atol=1e-08):
    """Use high precision for cumsum and check that final value matches sum.

    Parameters
    ----------
    arr : array-like
        To be cumulatively summed as flat
    axis : int, optional
        Axis along which the cumulative sum is computed.
        The default (None) is to compute the cumsum over the flattened array.
    rtol : float
        Relative tolerance, see ``np.allclose``
    atol : float
        Absolute tolerance, see ``np.allclose``
    """
    out = np.cumsum(arr, axis=axis, dtype=np.float64)
    expected = np.sum(arr, axis=axis, dtype=np.float64)
    if not np.all(
        np.isclose(
            out.take(-1, axis=axis), expected, rtol=rtol, atol=atol, equal_nan=True
        )
    ):
        warnings.warn(
            "cumsum was found to be unstable: "
            "its last element does not correspond to sum",
            RuntimeWarning,
        )
    return out


###############################################################################
# From nilearn


def _crop_colorbar(cbar, cbar_vmin, cbar_vmax):
    """Crop a colorbar to show from cbar_vmin to cbar_vmax.

    Used when symmetric_cbar=False is used.
    """
    if (cbar_vmin is None) and (cbar_vmax is None):
        return
    cbar_tick_locs = cbar.locator.locs
    if cbar_vmax is None:
        cbar_vmax = cbar_tick_locs.max()
    if cbar_vmin is None:
        cbar_vmin = cbar_tick_locs.min()
    new_tick_locs = np.linspace(cbar_vmin, cbar_vmax, len(cbar_tick_locs))

    cbar.ax.set_ylim(cbar_vmin, cbar_vmax)
    X = cbar._mesh()[0]
    X = np.array([X[0], X[-1]])
    Y = np.array([[cbar_vmin, cbar_vmin], [cbar_vmax, cbar_vmax]])
    N = X.shape[0]
    ii = [0, 1, N - 2, N - 1, 2 * N - 1, 2 * N - 2, N + 1, N, 0]
    x = X.T.reshape(-1)[ii]
    y = Y.T.reshape(-1)[ii]
    xy = (
        np.column_stack([y, x])
        if cbar.orientation == "horizontal"
        else np.column_stack([x, y])
    )
    cbar.outline.set_xy(xy)

    cbar.set_ticks(new_tick_locs)
    cbar.update_ticks()


###############################################################################
# Numba (optional requirement)

# Here we choose different defaults to speed things up by default
try:
    import numba

    if _compare_version(numba.__version__, "<", "0.53.1"):
        raise ImportError
    prange = numba.prange

    def jit(nopython=True, nogil=True, fastmath=True, cache=True, **kwargs):  # noqa
        return numba.jit(
            nopython=nopython, nogil=nogil, fastmath=fastmath, cache=cache, **kwargs
        )

except Exception:  # could be ImportError, SystemError, etc.
    has_numba = False
else:
    has_numba = os.getenv("MNE_USE_NUMBA", "true").lower() == "true"


if not has_numba:

    def jit(**kwargs):  # noqa
        def _jit(func):
            return func

        return _jit

    prange = range
    bincount = np.bincount

else:

    @jit()
    def bincount(x, weights, minlength):  # noqa: D103
        out = np.zeros(minlength)
        for idx, w in zip(x, weights):
            out[idx] += w
        return out


###############################################################################
# Matplotlib


# workaround: plt.close() doesn't spawn close_event on Agg backend
# https://github.com/matplotlib/matplotlib/issues/18609
def _close_event(fig):
    """Force calling of the MPL figure close event."""
    from matplotlib import backend_bases

    from .utils import logger

    try:
        fig.canvas.callbacks.process(
            "close_event",
            backend_bases.CloseEvent(name="close_event", canvas=fig.canvas),
        )
        logger.debug(f"Called {fig!r}.canvas.close_event()")
    except ValueError:  # old mpl with Qt
        logger.debug(f"Calling {fig!r}.canvas.close_event() failed")
        pass  # pragma: no cover


###############################################################################
# SciPy 1.14+ minimum_phase half=True option


def minimum_phase(h, method="homomorphic", n_fft=None, *, half=True):
    """Wrap scipy.signal.minimum_phase with half option."""
    # Can be removed once
    from scipy.fft import fft, ifft
    from scipy.signal import minimum_phase as sp_minimum_phase

    assert isinstance(method, str) and method == "homomorphic"

    if "half" in inspect.getfullargspec(sp_minimum_phase).kwonlyargs:
        return sp_minimum_phase(h, method=method, n_fft=n_fft, half=half)
    h = np.asarray(h)
    if np.iscomplexobj(h):
        raise ValueError("Complex filters not supported")
    if h.ndim != 1 or h.size <= 2:
        raise ValueError("h must be 1-D and at least 2 samples long")
    n_half = len(h) // 2
    if not np.allclose(h[-n_half:][::-1], h[:n_half]):
        warnings.warn(
            "h does not appear to by symmetric, conversion may fail",
            RuntimeWarning,
            stacklevel=2,
        )
    if n_fft is None:
        n_fft = 2 ** int(np.ceil(np.log2(2 * (len(h) - 1) / 0.01)))
    n_fft = int(n_fft)
    if n_fft < len(h):
        raise ValueError(f"n_fft must be at least len(h)=={len(h)}")

    # zero-pad; calculate the DFT
    h_temp = np.abs(fft(h, n_fft))
    # take 0.25*log(|H|**2) = 0.5*log(|H|)
    h_temp += 1e-7 * h_temp[h_temp > 0].min()  # don't let log blow up
    np.log(h_temp, out=h_temp)
    if half:  # halving of magnitude spectrum optional
        h_temp *= 0.5
    # IDFT
    h_temp = ifft(h_temp).real
    # multiply pointwise by the homomorphic filter
    # lmin[n] = 2u[n] - d[n]
    # i.e., double the positive frequencies and zero out the negative ones;
    # Oppenheim+Shafer 3rd ed p991 eq13.42b and p1004 fig13.7
    win = np.zeros(n_fft)
    win[0] = 1
    stop = n_fft // 2
    win[1:stop] = 2
    if n_fft % 2:
        win[stop] = 1
    h_temp *= win
    h_temp = ifft(np.exp(fft(h_temp)))
    h_minimum = h_temp.real

    n_out = (n_half + len(h) % 2) if half else len(h)
    return h_minimum[:n_out]



__all__ = [
    "ArgvSetter",
    "Bunch",
    "BunchConst",
    "BunchConstNamed",
    "ClosingStringIO",
    "ExtendedTimeMixin",
    "GetEpochsMixin",
    "ProgressBar",
    "SizeMixin",
    "TimeMixin",
    "_DefaultEventParser",
    "_PCA",
    "_ReuseCycle",
    "_TempDir",
    "_apply_scaling_array",
    "_apply_scaling_cov",
    "_arange_div",
    "_array_equal_nan",
    "_array_repr",
    "_assert_no_instances",
    "_auto_weakref",
    "_build_data_frame",
    "_check_all_same_channel_names",
    "_check_ch_locs",
    "_check_channels_spatial_filter",
    "_check_combine",
    "_check_compensation_grade",
    "_check_decim",
    "_check_depth",
    "_check_dict_keys",
    "_check_dt",
    "_check_edfio_installed",
    "_check_eeglabio_installed",
    "_check_event_id",
    "_check_fname",
    "_check_freesurfer_home",
    "_check_head_radius",
    "_check_if_nan",
    "_check_info_inv",
    "_check_integer_or_list",
    "_check_method_kwargs",
    "_check_on_missing",
    "_check_one_ch_type",
    "_check_option",
    "_check_pandas_index_arguments",
    "_check_pandas_installed",
    "_check_preload",
    "_check_pybv_installed",
    "_check_pymatreader_installed",
    "_check_qt_version",
    "_check_range",
    "_check_rank",
    "_check_sphere",
    "_check_src_normal",
    "_check_stc_units",
    "_check_subject",
    "_check_time_format",
    "_clean_names",
    "_click_ch_name",
    "_compute_row_norms",
    "_convert_times",
    "_custom_lru_cache",
    "_doc_special_members",
    "_date_to_julian",
    "_dt_to_stamp",
    "_empty_hash",
    "_ensure_events",
    "_ensure_int",
    "_explain_exception",
    "_file_like",
    "_freq_mask",
    "_gen_events",
    "_get_argvalues",
    "_get_blas_funcs",
    "_get_call_line",
    "_get_extra_data_path",
    "_get_inst_data",
    "_get_numpy_libs",
    "_get_root_dir",
    "_get_stim_channel",
    "_hashable_ndarray",
    "_import_h5io_funcs",
    "_import_h5py",
    "_import_nibabel",
    "_import_pymatreader_funcs",
    "_is_numeric",
    "_julian_to_date",
    "_mask_to_onsets_offsets",
    "_on_missing",
    "_parse_verbose",
    "_path_like",
    "_pl",
    "_prepare_read_metadata",
    "_prepare_write_metadata",
    "_raw_annot",
    "_record_warnings",
    "_reg_pinv",
    "_reject_data_segments",
    "_repeated_svd",
    "_replace_md5",
    "_require_version",
    "_resource_path",
    "_safe_input",
    "_scale_dataframe_data",
    "_scaled_array",
    "_set_pandas_dtype",
    "_soft_import",
    "_stamp_to_dt",
    "_suggest",
    "_svd_lwork",
    "_sym_mat_pow",
    "_time_mask",
    "_to_rgb",
    "_undo_scaling_array",
    "_undo_scaling_cov",
    "_url_to_local_path",
    "_validate_type",
    "_verbose_safe_false",
    "array_split_idx",
    "assert_and_remove_boundary_annot",
    "assert_dig_allclose",
    "assert_meg_snr",
    "assert_object_equal",
    "assert_snr",
    "assert_stcs_equal",
    "buggy_mkl_svd",
    "catch_logging",
    "check_fname",
    "check_random_state",
    "check_version",
    "compute_corr",
    "copy_doc",
    "copy_function_doc_to_method_doc",
    "create_slices",
    "deprecated",
    "deprecated_alias",
    "eigh",
    "fill_doc",
    "filter_out_warnings",
    "get_config",
    "get_config_path",
    "get_subjects_dir",
    "grand_average",
    "has_freesurfer",
    "has_mne_c",
    "hashfunc",
    "int_like",
    "legacy",
    "linkcode_resolve",
    "logger",
    "object_diff",
    "object_hash",
    "object_size",
    "open_docs",
    "path_like",
    "pformat",
    "pinv",
    "pinvh",
    "random_permutation",
    "repr_html",
    "requires_freesurfer",
    "requires_good_network",
    "requires_mne",
    "requires_mne_mark",
    "requires_openmeeg_mark",
    "run_command_if_main",
    "run_subprocess",
    "running_subprocess",
    "set_cache_dir",
    "set_config",
    "set_log_file",
    "set_log_level",
    "set_memmap_min_size",
    "sizeof_fmt",
    "split_list",
    "sqrtm_sym",
    "sum_squared",
    "sys_info",
    "use_log_level",
    "verbose",
    "warn",
    "wrapped_stdout",
]
from ._bunch import Bunch, BunchConst, BunchConstNamed
from ._logging import (
    ClosingStringIO,
    _get_call_line,
    _parse_verbose,
    _record_warnings,
    _verbose_safe_false,
    catch_logging,
    filter_out_warnings,
    logger,
    set_log_file,
    set_log_level,
    use_log_level,
    verbose,
    warn,
    wrapped_stdout,
)
from ._testing import (
    ArgvSetter,
    _click_ch_name,
    _raw_annot,
    _TempDir,
    assert_and_remove_boundary_annot,
    assert_dig_allclose,
    assert_meg_snr,
    assert_object_equal,
    assert_snr,
    assert_stcs_equal,
    buggy_mkl_svd,
    has_freesurfer,
    has_mne_c,
    requires_freesurfer,
    requires_good_network,
    requires_mne,
    requires_mne_mark,
    requires_openmeeg_mark,
    run_command_if_main,
)
from .check import (
    _check_all_same_channel_names,
    _check_ch_locs,
    _check_channels_spatial_filter,
    _check_combine,
    _check_compensation_grade,
    _check_depth,
    _check_dict_keys,
    _check_edfio_installed,
    _check_eeglabio_installed,
    _check_event_id,
    _check_fname,
    _check_freesurfer_home,
    _check_head_radius,
    _check_if_nan,
    _check_info_inv,
    _check_integer_or_list,
    _check_method_kwargs,
    _check_on_missing,
    _check_one_ch_type,
    _check_option,
    _check_pandas_index_arguments,
    _check_pandas_installed,
    _check_preload,
    _check_pybv_installed,
    _check_pymatreader_installed,
    _check_qt_version,
    _check_range,
    _check_rank,
    _check_sphere,
    _check_src_normal,
    _check_stc_units,
    _check_subject,
    _check_time_format,
    _ensure_events,
    _ensure_int,
    _import_h5io_funcs,
    _import_h5py,
    _import_nibabel,
    _import_pymatreader_funcs,
    _is_numeric,
    _on_missing,
    _path_like,
    _require_version,
    _safe_input,
    _soft_import,
    _suggest,
    _to_rgb,
    _validate_type,
    check_fname,
    check_random_state,
    check_version,
    int_like,
    path_like,
)
from .config import (
    _get_extra_data_path,
    _get_numpy_libs,
    _get_root_dir,
    _get_stim_channel,
    get_config,
    get_config_path,
    get_subjects_dir,
    set_cache_dir,
    set_config,
    set_memmap_min_size,
    sys_info,
)
from .dataframe import (
    _build_data_frame,
    _convert_times,
    _scale_dataframe_data,
    _set_pandas_dtype,
)
from .docs import (
    _doc_special_members,
    copy_doc,
    copy_function_doc_to_method_doc,
    deprecated,
    deprecated_alias,
    fill_doc,
    legacy,
    linkcode_resolve,
    open_docs,
)
from .fetching import _url_to_local_path
from .linalg import (
    _get_blas_funcs,
    _repeated_svd,
    _svd_lwork,
    _sym_mat_pow,
    eigh,
    pinv,
    pinvh,
    sqrtm_sym,
)
from .misc import (
    _assert_no_instances,
    _auto_weakref,
    _clean_names,
    _DefaultEventParser,
    _empty_hash,
    _explain_exception,
    _file_like,
    _get_argvalues,
    _pl,
    _resource_path,
    pformat,
    repr_html,
    run_subprocess,
    running_subprocess,
    sizeof_fmt,
)
from .mixin import (
    ExtendedTimeMixin,
    GetEpochsMixin,
    SizeMixin,
    TimeMixin,
    _check_decim,
    _prepare_read_metadata,
    _prepare_write_metadata,
)
from .numerics import (
    _PCA,
    _apply_scaling_array,
    _apply_scaling_cov,
    _arange_div,
    _array_equal_nan,
    _array_repr,
    _check_dt,
    _compute_row_norms,
    _custom_lru_cache,
    _date_to_julian,
    _dt_to_stamp,
    _freq_mask,
    _gen_events,
    _get_inst_data,
    _hashable_ndarray,
    _julian_to_date,
    _mask_to_onsets_offsets,
    _reg_pinv,
    _reject_data_segments,
    _replace_md5,
    _ReuseCycle,
    _scaled_array,
    _stamp_to_dt,
    _time_mask,
    _undo_scaling_array,
    _undo_scaling_cov,
    array_split_idx,
    compute_corr,
    create_slices,
    grand_average,
    hashfunc,
    object_diff,
    object_hash,
    object_size,
    random_permutation,
    split_list,
    sum_squared,
)
from .progressbar import ProgressBar



# Authors: The MNE-Python contributors.
# License: BSD-3-Clause
# Copyright the MNE-Python contributors.

import numpy as np
from numpy.polynomial.legendre import legval
from scipy.interpolate import RectBivariateSpline
from scipy.linalg import pinv
from scipy.spatial.distance import pdist, squareform

from .._fiff.meas_info import _simplify_info
from .._fiff.pick import pick_channels, pick_info, pick_types
from ..surface import _normalize_vectors
from ..utils import _validate_type, logger, verbose, warn


def _calc_h(cosang, stiffness=4, n_legendre_terms=50):
    """Calculate spherical spline h function between points on a sphere.

    Parameters
    ----------
    cosang : array-like | float
        cosine of angles between pairs of points on a spherical surface. This
        is equivalent to the dot product of unit vectors.
    stiffness : float
        stiffnes of the spline. Also referred to as ``m``.
    n_legendre_terms : int
        number of Legendre terms to evaluate.
    """
    factors = [
        (2 * n + 1) / (n ** (stiffness - 1) * (n + 1) ** (stiffness - 1) * 4 * np.pi)
        for n in range(1, n_legendre_terms + 1)
    ]
    return legval(cosang, [0] + factors)


def _calc_g(cosang, stiffness=4, n_legendre_terms=50):
    """Calculate spherical spline g function between points on a sphere.

    Parameters
    ----------
    cosang : array-like of float, shape(n_channels, n_channels)
        cosine of angles between pairs of points on a spherical surface. This
        is equivalent to the dot product of unit vectors.
    stiffness : float
        stiffness of the spline.
    n_legendre_terms : int
        number of Legendre terms to evaluate.

    Returns
    -------
    G : np.ndrarray of float, shape(n_channels, n_channels)
        The G matrix.
    """
    factors = [
        (2 * n + 1) / (n**stiffness * (n + 1) ** stiffness * 4 * np.pi)
        for n in range(1, n_legendre_terms + 1)
    ]
    return legval(cosang, [0] + factors)


def _make_interpolation_matrix(pos_from, pos_to, alpha=1e-5):
    """Compute interpolation matrix based on spherical splines.

    Implementation based on [1]

    Parameters
    ----------
    pos_from : np.ndarray of float, shape(n_good_sensors, 3)
        The positions to interpolate from.
    pos_to : np.ndarray of float, shape(n_bad_sensors, 3)
        The positions to interpolate.
    alpha : float
        Regularization parameter. Defaults to 1e-5.

    Returns
    -------
    interpolation : np.ndarray of float, shape(len(pos_from), len(pos_to))
        The interpolation matrix that maps good signals to the location
        of bad signals.

    References
    ----------
    [1] Perrin, F., Pernier, J., Bertrand, O. and Echallier, JF. (1989).
        Spherical splines for scalp potential and current density mapping.
        Electroencephalography Clinical Neurophysiology, Feb; 72(2):184-7.
    """
    pos_from = pos_from.copy()
    pos_to = pos_to.copy()
    n_from = pos_from.shape[0]
    n_to = pos_to.shape[0]

    # normalize sensor positions to sphere
    _normalize_vectors(pos_from)
    _normalize_vectors(pos_to)

    # cosine angles between source positions
    cosang_from = pos_from.dot(pos_from.T)
    cosang_to_from = pos_to.dot(pos_from.T)
    G_from = _calc_g(cosang_from)
    G_to_from = _calc_g(cosang_to_from)
    assert G_from.shape == (n_from, n_from)
    assert G_to_from.shape == (n_to, n_from)

    if alpha is not None:
        G_from.flat[:: len(G_from) + 1] += alpha

    C = np.vstack(
        [
            np.hstack([G_from, np.ones((n_from, 1))]),
            np.hstack([np.ones((1, n_from)), [[0]]]),
        ]
    )
    C_inv = pinv(C)

    interpolation = np.hstack([G_to_from, np.ones((n_to, 1))]) @ C_inv[:, :-1]
    assert interpolation.shape == (n_to, n_from)
    return interpolation


def _do_interp_dots(inst, interpolation, goods_idx, bads_idx):
    """Dot product of channel mapping matrix to channel data."""
    from ..epochs import BaseEpochs
    from ..evoked import Evoked
    from ..io import BaseRaw

    _validate_type(inst, (BaseRaw, BaseEpochs, Evoked), "inst")
    inst._data[..., bads_idx, :] = np.matmul(
        interpolation, inst._data[..., goods_idx, :]
    )


@verbose
def _interpolate_bads_eeg(inst, origin, exclude=None, ecog=False, verbose=None):
    if exclude is None:
        exclude = list()
    bads_idx = np.zeros(len(inst.ch_names), dtype=bool)
    goods_idx = np.zeros(len(inst.ch_names), dtype=bool)

    picks = pick_types(inst.info, meg=False, eeg=not ecog, ecog=ecog, exclude=exclude)
    inst.info._check_consistency()
    bads_idx[picks] = [inst.ch_names[ch] in inst.info["bads"] for ch in picks]

    if len(picks) == 0 or bads_idx.sum() == 0:
        return

    goods_idx[picks] = True
    goods_idx[bads_idx] = False

    pos = inst._get_channel_positions(picks)

    # Make sure only EEG are used
    bads_idx_pos = bads_idx[picks]
    goods_idx_pos = goods_idx[picks]

    # test spherical fit
    distance = np.linalg.norm(pos - origin, axis=-1)
    distance = np.mean(distance / np.mean(distance))
    if np.abs(1.0 - distance) > 0.1:
        warn(
            "Your spherical fit is poor, interpolation results are "
            "likely to be inaccurate."
        )

    pos_good = pos[goods_idx_pos] - origin
    pos_bad = pos[bads_idx_pos] - origin
    logger.info(f"Computing interpolation matrix from {len(pos_good)} sensor positions")
    interpolation = _make_interpolation_matrix(pos_good, pos_bad)

    logger.info(f"Interpolating {len(pos_bad)} sensors")
    _do_interp_dots(inst, interpolation, goods_idx, bads_idx)


@verbose
def _interpolate_bads_ecog(inst, origin, exclude=None, verbose=None):
    _interpolate_bads_eeg(inst, origin, exclude=exclude, ecog=True, verbose=verbose)


def _interpolate_bads_meg(
    inst, mode="accurate", origin=(0.0, 0.0, 0.04), verbose=None, ref_meg=False
):
    return _interpolate_bads_meeg(
        inst, mode, origin, ref_meg=ref_meg, eeg=False, verbose=verbose
    )


@verbose
def _interpolate_bads_nan(
    inst,
    ch_type,
    ref_meg=False,
    exclude=(),
    *,
    verbose=None,
):
    info = _simplify_info(inst.info)
    picks_type = pick_types(info, ref_meg=ref_meg, exclude=exclude, **{ch_type: True})
    use_ch_names = [inst.info["ch_names"][p] for p in picks_type]
    bads_type = [ch for ch in inst.info["bads"] if ch in use_ch_names]
    if len(bads_type) == 0 or len(picks_type) == 0:
        return
    # select the bad channels to be interpolated
    picks_bad = pick_channels(inst.info["ch_names"], bads_type, exclude=[])
    inst._data[..., picks_bad, :] = np.nan


@verbose
def _interpolate_bads_meeg(
    inst,
    mode="accurate",
    origin=(0.0, 0.0, 0.04),
    meg=True,
    eeg=True,
    ref_meg=False,
    exclude=(),
    *,
    method=None,
    verbose=None,
):
    from ..forward import _map_meg_or_eeg_channels

    if method is None:
        method = {"meg": "MNE", "eeg": "MNE"}
    bools = dict(meg=meg, eeg=eeg)
    info = _simplify_info(inst.info)
    for ch_type, do in bools.items():
        if not do:
            continue
        kw = dict(meg=False, eeg=False)
        kw[ch_type] = True
        picks_type = pick_types(info, ref_meg=ref_meg, exclude=exclude, **kw)
        picks_good = pick_types(info, ref_meg=ref_meg, exclude="bads", **kw)
        use_ch_names = [inst.info["ch_names"][p] for p in picks_type]
        bads_type = [ch for ch in inst.info["bads"] if ch in use_ch_names]
        if len(bads_type) == 0 or len(picks_type) == 0:
            continue
        # select the bad channels to be interpolated
        picks_bad = pick_channels(inst.info["ch_names"], bads_type, exclude=[])

        # do MNE based interpolation
        if ch_type == "eeg":
            picks_to = picks_type
            bad_sel = np.isin(picks_type, picks_bad)
        else:
            picks_to = picks_bad
            bad_sel = slice(None)
        info_from = pick_info(inst.info, picks_good)
        info_to = pick_info(inst.info, picks_to)
        mapping = _map_meg_or_eeg_channels(info_from, info_to, mode=mode, origin=origin)
        mapping = mapping[bad_sel]
        _do_interp_dots(inst, mapping, picks_good, picks_bad)


@verbose
def _interpolate_bads_nirs(inst, exclude=(), verbose=None):
    from mne.preprocessing.nirs import _validate_nirs_info

    if len(pick_types(inst.info, fnirs=True, exclude=())) == 0:
        return

    # Returns pick of all nirs and ensures channels are correctly ordered
    picks_nirs = _validate_nirs_info(inst.info)
    nirs_ch_names = [inst.info["ch_names"][p] for p in picks_nirs]
    nirs_ch_names = [ch for ch in nirs_ch_names if ch not in exclude]
    bads_nirs = [ch for ch in inst.info["bads"] if ch in nirs_ch_names]
    if len(bads_nirs) == 0:
        return
    picks_bad = pick_channels(inst.info["ch_names"], bads_nirs, exclude=[])
    bads_mask = [p in picks_bad for p in picks_nirs]

    chs = [inst.info["chs"][i] for i in picks_nirs]
    locs3d = np.array([ch["loc"][:3] for ch in chs])

    dist = pdist(locs3d)
    dist = squareform(dist)

    for bad in picks_bad:
        dists_to_bad = dist[bad]
        # Ignore distances to self
        dists_to_bad[dists_to_bad == 0] = np.inf
        # Ignore distances to other bad channels
        dists_to_bad[bads_mask] = np.inf
        # Find closest remaining channels for same frequency
        closest_idx = np.argmin(dists_to_bad) + (bad % 2)
        inst._data[bad] = inst._data[closest_idx]

    # TODO: this seems like a bug because it does not respect reset_bads
    inst.info["bads"] = [ch for ch in inst.info["bads"] if ch in exclude]

    return inst


def _find_seeg_electrode_shaft(pos, tol_shaft=0.002, tol_spacing=1):
    # 1) find nearest neighbor to define the electrode shaft line
    # 2) find all contacts on the same line
    # 3) remove contacts with large distances

    dist = squareform(pdist(pos))
    np.fill_diagonal(dist, np.inf)

    shafts = list()
    shaft_ts = list()
    for i, n1 in enumerate(pos):
        if any([i in shaft for shaft in shafts]):
            continue
        n2 = pos[np.argmin(dist[i])]  # 1
        # https://mathworld.wolfram.com/Point-LineDistance3-Dimensional.html
        shaft_dists = np.linalg.norm(
            np.cross((pos - n1), (pos - n2)), axis=1
        ) / np.linalg.norm(n2 - n1)
        shaft = np.where(shaft_dists < tol_shaft)[0]  # 2
        shaft_prev = None
        for _ in range(10):  # avoid potential cycles
            if np.array_equal(shaft, shaft_prev):
                break
            shaft_prev = shaft
            # compute median shaft line
            v = np.median(
                [
                    pos[i] - pos[j]
                    for idx, i in enumerate(shaft)
                    for j in shaft[idx + 1 :]
                ],
                axis=0,
            )
            c = np.median(pos[shaft], axis=0)
            # recompute distances
            shaft_dists = np.linalg.norm(
                np.cross((pos - c), (pos - c + v)), axis=1
            ) / np.linalg.norm(v)
            shaft = np.where(shaft_dists < tol_shaft)[0]
        ts = np.array([np.dot(c - n0, v) / np.linalg.norm(v) ** 2 for n0 in pos[shaft]])
        shaft_order = np.argsort(ts)
        shaft = shaft[shaft_order]
        ts = ts[shaft_order]

        # only include the largest group with spacing with the error tolerance
        # avoid interpolating across spans between contacts
        t_diffs = np.diff(ts)
        t_diff_med = np.median(t_diffs)
        spacing_errors = (t_diffs - t_diff_med) / t_diff_med
        groups = list()
        group = [shaft[0]]
        for j in range(len(shaft) - 1):
            if spacing_errors[j] > tol_spacing:
                groups.append(group)
                group = [shaft[j + 1]]
            else:
                group.append(shaft[j + 1])
        groups.append(group)
        group = [group for group in groups if i in group][0]
        ts = ts[np.isin(shaft, group)]
        shaft = np.array(group, dtype=int)

        shafts.append(shaft)
        shaft_ts.append(ts)
    return shafts, shaft_ts


@verbose
def _interpolate_bads_seeg(
    inst, exclude=None, tol_shaft=0.002, tol_spacing=1, verbose=None
):
    if exclude is None:
        exclude = list()
    picks = pick_types(inst.info, meg=False, seeg=True, exclude=exclude)
    inst.info._check_consistency()
    bads_idx = np.isin(np.array(inst.ch_names)[picks], inst.info["bads"])

    if len(picks) == 0 or bads_idx.sum() == 0:
        return

    pos = inst._get_channel_positions(picks)

    # Make sure only sEEG are used
    bads_idx_pos = bads_idx[picks]

    shafts, shaft_ts = _find_seeg_electrode_shaft(
        pos, tol_shaft=tol_shaft, tol_spacing=tol_spacing
    )

    # interpolate the bad contacts
    picks_bad = list(np.where(bads_idx_pos)[0])
    for shaft, ts in zip(shafts, shaft_ts):
        bads_shaft = np.array([idx for idx in picks_bad if idx in shaft])
        if bads_shaft.size == 0:
            continue
        goods_shaft = shaft[np.isin(shaft, bads_shaft, invert=True)]
        if goods_shaft.size < 4:  # cubic spline requires 3 channels
            msg = "No shaft" if shaft.size < 4 else "Not enough good channels"
            no_shaft_chs = " and ".join(np.array(inst.ch_names)[bads_shaft])
            raise RuntimeError(
                f"{msg} found in a line with {no_shaft_chs} "
                "at least 3 good channels on the same line "
                f"are required for interpolation, {goods_shaft.size} found. "
                f"Dropping {no_shaft_chs} is recommended."
            )
        logger.debug(
            f"Interpolating {np.array(inst.ch_names)[bads_shaft]} using "
            f"data from {np.array(inst.ch_names)[goods_shaft]}"
        )
        bads_shaft_idx = np.where(np.isin(shaft, bads_shaft))[0]
        goods_shaft_idx = np.where(~np.isin(shaft, bads_shaft))[0]

        z = inst._data[..., goods_shaft, :]
        is_epochs = z.ndim == 3
        if is_epochs:
            z = z.swapaxes(0, 1)
            z = z.reshape(z.shape[0], -1)
        y = np.arange(z.shape[-1])
        out = RectBivariateSpline(x=ts[goods_shaft_idx], y=y, z=z)(
            x=ts[bads_shaft_idx], y=y
        )
        if is_epochs:
            out = out.reshape(bads_shaft.size, inst._data.shape[0], -1)
            out = out.swapaxes(0, 1)
        inst._data[..., bads_shaft, :] = out
